# Vendored from
# https://github.com/matplotlib/matplotlib/blob/6166e5b990b32c7f46847782c0354653e483560d/lib/matplotlib/_color_data.py

BASE_COLORS = {
    "b": (0, 0, 1),  # blue
    "g": (0, 0.5, 0),  # green
    "r": (1, 0, 0),  # red
    "c": (0, 0.75, 0.75),  # cyan
    "m": (0.75, 0, 0.75),  # magenta
    "y": (0.75, 0.75, 0),  # yellow
    "k": (0, 0, 0),  # black
    "w": (1, 1, 1),  # white
}


# These colors are from Tableau
TABLEAU_COLORS = {
    "tab:blue": "#1f77b4",
    "tab:orange": "#ff7f0e",
    "tab:green": "#2ca02c",
    "tab:red": "#d62728",
    "tab:purple": "#9467bd",
    "tab:brown": "#8c564b",
    "tab:pink": "#e377c2",
    "tab:gray": "#7f7f7f",
    "tab:olive": "#bcbd22",
    "tab:cyan": "#17becf",
}


# This mapping of color names -> hex values is taken from
# a survey run by Randall Munroe see:
# https://blog.xkcd.com/2010/05/03/color-survey-results/
# for more details.  The results are hosted at
# https://xkcd.com/color/rgb/
# and also available as a text file at
# https://xkcd.com/color/rgb.txt
#
# License: https://creativecommons.org/publicdomain/zero/1.0/
XKCD_COLORS = {
    "cloudy blue": "#acc2d9",
    "dark pastel green": "#56ae57",
    "dust": "#b2996e",
    "electric lime": "#a8ff04",
    "fresh green": "#69d84f",
    "light eggplant": "#894585",
    "nasty green": "#70b23f",
    "really light blue": "#d4ffff",
    "tea": "#65ab7c",
    "warm purple": "#952e8f",
    "yellowish tan": "#fcfc81",
    "cement": "#a5a391",
    "dark grass green": "#388004",
    "dusty teal": "#4c9085",
    "grey teal": "#5e9b8a",
    "macaroni and cheese": "#efb435",
    "pinkish tan": "#d99b82",
    "spruce": "#0a5f38",
    "strong blue": "#0c06f7",
    "toxic green": "#61de2a",
    "windows blue": "#3778bf",
    "blue blue": "#2242c7",
    "blue with a hint of purple": "#533cc6",
    "booger": "#9bb53c",
    "bright sea green": "#05ffa6",
    "dark green blue": "#1f6357",
    "deep turquoise": "#017374",
    "green teal": "#0cb577",
    "strong pink": "#ff0789",
    "bland": "#afa88b",
    "deep aqua": "#08787f",
    "lavender pink": "#dd85d7",
    "light moss green": "#a6c875",
    "light seafoam green": "#a7ffb5",
    "olive yellow": "#c2b709",
    "pig pink": "#e78ea5",
    "deep lilac": "#966ebd",
    "desert": "#ccad60",
    "dusty lavender": "#ac86a8",
    "purpley grey": "#947e94",
    "purply": "#983fb2",
    "candy pink": "#ff63e9",
    "light pastel green": "#b2fba5",
    "boring green": "#63b365",
    "kiwi green": "#8ee53f",
    "light grey green": "#b7e1a1",
    "orange pink": "#ff6f52",
    "tea green": "#bdf8a3",
    "very light brown": "#d3b683",
    "egg shell": "#fffcc4",
    "eggplant purple": "#430541",
    "powder pink": "#ffb2d0",
    "reddish grey": "#997570",
    "baby shit brown": "#ad900d",
    "liliac": "#c48efd",
    "stormy blue": "#507b9c",
    "ugly brown": "#7d7103",
    "custard": "#fffd78",
    "darkish pink": "#da467d",
    "deep brown": "#410200",
    "greenish beige": "#c9d179",
    "manilla": "#fffa86",
    "off blue": "#5684ae",
    "battleship grey": "#6b7c85",
    "browny green": "#6f6c0a",
    "bruise": "#7e4071",
    "kelley green": "#009337",
    "sickly yellow": "#d0e429",
    "sunny yellow": "#fff917",
    "azul": "#1d5dec",
    "darkgreen": "#054907",
    "green/yellow": "#b5ce08",
    "lichen": "#8fb67b",
    "light light green": "#c8ffb0",
    "pale gold": "#fdde6c",
    "sun yellow": "#ffdf22",
    "tan green": "#a9be70",
    "burple": "#6832e3",
    "butterscotch": "#fdb147",
    "toupe": "#c7ac7d",
    "dark cream": "#fff39a",
    "indian red": "#850e04",
    "light lavendar": "#efc0fe",
    "poison green": "#40fd14",
    "baby puke green": "#b6c406",
    "bright yellow green": "#9dff00",
    "charcoal grey": "#3c4142",
    "squash": "#f2ab15",
    "cinnamon": "#ac4f06",
    "light pea green": "#c4fe82",
    "radioactive green": "#2cfa1f",
    "raw sienna": "#9a6200",
    "baby purple": "#ca9bf7",
    "cocoa": "#875f42",
    "light royal blue": "#3a2efe",
    "orangeish": "#fd8d49",
    "rust brown": "#8b3103",
    "sand brown": "#cba560",
    "swamp": "#698339",
    "tealish green": "#0cdc73",
    "burnt siena": "#b75203",
    "camo": "#7f8f4e",
    "dusk blue": "#26538d",
    "fern": "#63a950",
    "old rose": "#c87f89",
    "pale light green": "#b1fc99",
    "peachy pink": "#ff9a8a",
    "rosy pink": "#f6688e",
    "light bluish green": "#76fda8",
    "light bright green": "#53fe5c",
    "light neon green": "#4efd54",
    "light seafoam": "#a0febf",
    "tiffany blue": "#7bf2da",
    "washed out green": "#bcf5a6",
    "browny orange": "#ca6b02",
    "nice blue": "#107ab0",
    "sapphire": "#2138ab",
    "greyish teal": "#719f91",
    "orangey yellow": "#fdb915",
    "parchment": "#fefcaf",
    "straw": "#fcf679",
    "very dark brown": "#1d0200",
    "terracota": "#cb6843",
    "ugly blue": "#31668a",
    "clear blue": "#247afd",
    "creme": "#ffffb6",
    "foam green": "#90fda9",
    "grey/green": "#86a17d",
    "light gold": "#fddc5c",
    "seafoam blue": "#78d1b6",
    "topaz": "#13bbaf",
    "violet pink": "#fb5ffc",
    "wintergreen": "#20f986",
    "yellow tan": "#ffe36e",
    "dark fuchsia": "#9d0759",
    "indigo blue": "#3a18b1",
    "light yellowish green": "#c2ff89",
    "pale magenta": "#d767ad",
    "rich purple": "#720058",
    "sunflower yellow": "#ffda03",
    "green/blue": "#01c08d",
    "leather": "#ac7434",
    "racing green": "#014600",
    "vivid purple": "#9900fa",
    "dark royal blue": "#02066f",
    "hazel": "#8e7618",
    "muted pink": "#d1768f",
    "booger green": "#96b403",
    "canary": "#fdff63",
    "cool grey": "#95a3a6",
    "dark taupe": "#7f684e",
    "darkish purple": "#751973",
    "true green": "#089404",
    "coral pink": "#ff6163",
    "dark sage": "#598556",
    "dark slate blue": "#214761",
    "flat blue": "#3c73a8",
    "mushroom": "#ba9e88",
    "rich blue": "#021bf9",
    "dirty purple": "#734a65",
    "greenblue": "#23c48b",
    "icky green": "#8fae22",
    "light khaki": "#e6f2a2",
    "warm blue": "#4b57db",
    "dark hot pink": "#d90166",
    "deep sea blue": "#015482",
    "carmine": "#9d0216",
    "dark yellow green": "#728f02",
    "pale peach": "#ffe5ad",
    "plum purple": "#4e0550",
    "golden rod": "#f9bc08",
    "neon red": "#ff073a",
    "old pink": "#c77986",
    "very pale blue": "#d6fffe",
    "blood orange": "#fe4b03",
    "grapefruit": "#fd5956",
    "sand yellow": "#fce166",
    "clay brown": "#b2713d",
    "dark blue grey": "#1f3b4d",
    "flat green": "#699d4c",
    "light green blue": "#56fca2",
    "warm pink": "#fb5581",
    "dodger blue": "#3e82fc",
    "gross green": "#a0bf16",
    "ice": "#d6fffa",
    "metallic blue": "#4f738e",
    "pale salmon": "#ffb19a",
    "sap green": "#5c8b15",
    "algae": "#54ac68",
    "bluey grey": "#89a0b0",
    "greeny grey": "#7ea07a",
    "highlighter green": "#1bfc06",
    "light light blue": "#cafffb",
    "light mint": "#b6ffbb",
    "raw umber": "#a75e09",
    "vivid blue": "#152eff",
    "deep lavender": "#8d5eb7",
    "dull teal": "#5f9e8f",
    "light greenish blue": "#63f7b4",
    "mud green": "#606602",
    "pinky": "#fc86aa",
    "red wine": "#8c0034",
    "shit green": "#758000",
    "tan brown": "#ab7e4c",
    "darkblue": "#030764",
    "rosa": "#fe86a4",
    "lipstick": "#d5174e",
    "pale mauve": "#fed0fc",
    "claret": "#680018",
    "dandelion": "#fedf08",
    "orangered": "#fe420f",
    "poop green": "#6f7c00",
    "ruby": "#ca0147",
    "dark": "#1b2431",
    "greenish turquoise": "#00fbb0",
    "pastel red": "#db5856",
    "piss yellow": "#ddd618",
    "bright cyan": "#41fdfe",
    "dark coral": "#cf524e",
    "algae green": "#21c36f",
    "darkish red": "#a90308",
    "reddy brown": "#6e1005",
    "blush pink": "#fe828c",
    "camouflage green": "#4b6113",
    "lawn green": "#4da409",
    "putty": "#beae8a",
    "vibrant blue": "#0339f8",
    "dark sand": "#a88f59",
    "purple/blue": "#5d21d0",
    "saffron": "#feb209",
    "twilight": "#4e518b",
    "warm brown": "#964e02",
    "bluegrey": "#85a3b2",
    "bubble gum pink": "#ff69af",
    "duck egg blue": "#c3fbf4",
    "greenish cyan": "#2afeb7",
    "petrol": "#005f6a",
    "royal": "#0c1793",
    "butter": "#ffff81",
    "dusty orange": "#f0833a",
    "off yellow": "#f1f33f",
    "pale olive green": "#b1d27b",
    "orangish": "#fc824a",
    "leaf": "#71aa34",
    "light blue grey": "#b7c9e2",
    "dried blood": "#4b0101",
    "lightish purple": "#a552e6",
    "rusty red": "#af2f0d",
    "lavender blue": "#8b88f8",
    "light grass green": "#9af764",
    "light mint green": "#a6fbb2",
    "sunflower": "#ffc512",
    "velvet": "#750851",
    "brick orange": "#c14a09",
    "lightish red": "#fe2f4a",
    "pure blue": "#0203e2",
    "twilight blue": "#0a437a",
    "violet red": "#a50055",
    "yellowy brown": "#ae8b0c",
    "carnation": "#fd798f",
    "muddy yellow": "#bfac05",
    "dark seafoam green": "#3eaf76",
    "deep rose": "#c74767",
    "dusty red": "#b9484e",
    "grey/blue": "#647d8e",
    "lemon lime": "#bffe28",
    "purple/pink": "#d725de",
    "brown yellow": "#b29705",
    "purple brown": "#673a3f",
    "wisteria": "#a87dc2",
    "banana yellow": "#fafe4b",
    "lipstick red": "#c0022f",
    "water blue": "#0e87cc",
    "brown grey": "#8d8468",
    "vibrant purple": "#ad03de",
    "baby green": "#8cff9e",
    "barf green": "#94ac02",
    "eggshell blue": "#c4fff7",
    "sandy yellow": "#fdee73",
    "cool green": "#33b864",
    "pale": "#fff9d0",
    "blue/grey": "#758da3",
    "hot magenta": "#f504c9",
    "greyblue": "#77a1b5",
    "purpley": "#8756e4",
    "baby shit green": "#889717",
    "brownish pink": "#c27e79",
    "dark aquamarine": "#017371",
    "diarrhea": "#9f8303",
    "light mustard": "#f7d560",
    "pale sky blue": "#bdf6fe",
    "turtle green": "#75b84f",
    "bright olive": "#9cbb04",
    "dark grey blue": "#29465b",
    "greeny brown": "#696006",
    "lemon green": "#adf802",
    "light periwinkle": "#c1c6fc",
    "seaweed green": "#35ad6b",
    "sunshine yellow": "#fffd37",
    "ugly purple": "#a442a0",
    "medium pink": "#f36196",
    "puke brown": "#947706",
    "very light pink": "#fff4f2",
    "viridian": "#1e9167",
    "bile": "#b5c306",
    "faded yellow": "#feff7f",
    "very pale green": "#cffdbc",
    "vibrant green": "#0add08",
    "bright lime": "#87fd05",
    "spearmint": "#1ef876",
    "light aquamarine": "#7bfdc7",
    "light sage": "#bcecac",
    "yellowgreen": "#bbf90f",
    "baby poo": "#ab9004",
    "dark seafoam": "#1fb57a",
    "deep teal": "#00555a",
    "heather": "#a484ac",
    "rust orange": "#c45508",
    "dirty blue": "#3f829d",
    "fern green": "#548d44",
    "bright lilac": "#c95efb",
    "weird green": "#3ae57f",
    "peacock blue": "#016795",
    "avocado green": "#87a922",
    "faded orange": "#f0944d",
    "grape purple": "#5d1451",
    "hot green": "#25ff29",
    "lime yellow": "#d0fe1d",
    "mango": "#ffa62b",
    "shamrock": "#01b44c",
    "bubblegum": "#ff6cb5",
    "purplish brown": "#6b4247",
    "vomit yellow": "#c7c10c",
    "pale cyan": "#b7fffa",
    "key lime": "#aeff6e",
    "tomato red": "#ec2d01",
    "lightgreen": "#76ff7b",
    "merlot": "#730039",
    "night blue": "#040348",
    "purpleish pink": "#df4ec8",
    "apple": "#6ecb3c",
    "baby poop green": "#8f9805",
    "green apple": "#5edc1f",
    "heliotrope": "#d94ff5",
    "yellow/green": "#c8fd3d",
    "almost black": "#070d0d",
    "cool blue": "#4984b8",
    "leafy green": "#51b73b",
    "mustard brown": "#ac7e04",
    "dusk": "#4e5481",
    "dull brown": "#876e4b",
    "frog green": "#58bc08",
    "vivid green": "#2fef10",
    "bright light green": "#2dfe54",
    "fluro green": "#0aff02",
    "kiwi": "#9cef43",
    "seaweed": "#18d17b",
    "navy green": "#35530a",
    "ultramarine blue": "#1805db",
    "iris": "#6258c4",
    "pastel orange": "#ff964f",
    "yellowish orange": "#ffab0f",
    "perrywinkle": "#8f8ce7",
    "tealish": "#24bca8",
    "dark plum": "#3f012c",
    "pear": "#cbf85f",
    "pinkish orange": "#ff724c",
    "midnight purple": "#280137",
    "light urple": "#b36ff6",
    "dark mint": "#48c072",
    "greenish tan": "#bccb7a",
    "light burgundy": "#a8415b",
    "turquoise blue": "#06b1c4",
    "ugly pink": "#cd7584",
    "sandy": "#f1da7a",
    "electric pink": "#ff0490",
    "muted purple": "#805b87",
    "mid green": "#50a747",
    "greyish": "#a8a495",
    "neon yellow": "#cfff04",
    "banana": "#ffff7e",
    "carnation pink": "#ff7fa7",
    "tomato": "#ef4026",
    "sea": "#3c9992",
    "muddy brown": "#886806",
    "turquoise green": "#04f489",
    "buff": "#fef69e",
    "fawn": "#cfaf7b",
    "muted blue": "#3b719f",
    "pale rose": "#fdc1c5",
    "dark mint green": "#20c073",
    "amethyst": "#9b5fc0",
    "blue/green": "#0f9b8e",
    "chestnut": "#742802",
    "sick green": "#9db92c",
    "pea": "#a4bf20",
    "rusty orange": "#cd5909",
    "stone": "#ada587",
    "rose red": "#be013c",
    "pale aqua": "#b8ffeb",
    "deep orange": "#dc4d01",
    "earth": "#a2653e",
    "mossy green": "#638b27",
    "grassy green": "#419c03",
    "pale lime green": "#b1ff65",
    "light grey blue": "#9dbcd4",
    "pale grey": "#fdfdfe",
    "asparagus": "#77ab56",
    "blueberry": "#464196",
    "purple red": "#990147",
    "pale lime": "#befd73",
    "greenish teal": "#32bf84",
    "caramel": "#af6f09",
    "deep magenta": "#a0025c",
    "light peach": "#ffd8b1",
    "milk chocolate": "#7f4e1e",
    "ocher": "#bf9b0c",
    "off green": "#6ba353",
    "purply pink": "#f075e6",
    "lightblue": "#7bc8f6",
    "dusky blue": "#475f94",
    "golden": "#f5bf03",
    "light beige": "#fffeb6",
    "butter yellow": "#fffd74",
    "dusky purple": "#895b7b",
    "french blue": "#436bad",
    "ugly yellow": "#d0c101",
    "greeny yellow": "#c6f808",
    "orangish red": "#f43605",
    "shamrock green": "#02c14d",
    "orangish brown": "#b25f03",
    "tree green": "#2a7e19",
    "deep violet": "#490648",
    "gunmetal": "#536267",
    "blue/purple": "#5a06ef",
    "cherry": "#cf0234",
    "sandy brown": "#c4a661",
    "warm grey": "#978a84",
    "dark indigo": "#1f0954",
    "midnight": "#03012d",
    "bluey green": "#2bb179",
    "grey pink": "#c3909b",
    "soft purple": "#a66fb5",
    "blood": "#770001",
    "brown red": "#922b05",
    "medium grey": "#7d7f7c",
    "berry": "#990f4b",
    "poo": "#8f7303",
    "purpley pink": "#c83cb9",
    "light salmon": "#fea993",
    "snot": "#acbb0d",
    "easter purple": "#c071fe",
    "light yellow green": "#ccfd7f",
    "dark navy blue": "#00022e",
    "drab": "#828344",
    "light rose": "#ffc5cb",
    "rouge": "#ab1239",
    "purplish red": "#b0054b",
    "slime green": "#99cc04",
    "baby poop": "#937c00",
    "irish green": "#019529",
    "pink/purple": "#ef1de7",
    "dark navy": "#000435",
    "greeny blue": "#42b395",
    "light plum": "#9d5783",
    "pinkish grey": "#c8aca9",
    "dirty orange": "#c87606",
    "rust red": "#aa2704",
    "pale lilac": "#e4cbff",
    "orangey red": "#fa4224",
    "primary blue": "#0804f9",
    "kermit green": "#5cb200",
    "brownish purple": "#76424e",
    "murky green": "#6c7a0e",
    "wheat": "#fbdd7e",
    "very dark purple": "#2a0134",
    "bottle green": "#044a05",
    "watermelon": "#fd4659",
    "deep sky blue": "#0d75f8",
    "fire engine red": "#fe0002",
    "yellow ochre": "#cb9d06",
    "pumpkin orange": "#fb7d07",
    "pale olive": "#b9cc81",
    "light lilac": "#edc8ff",
    "lightish green": "#61e160",
    "carolina blue": "#8ab8fe",
    "mulberry": "#920a4e",
    "shocking pink": "#fe02a2",
    "auburn": "#9a3001",
    "bright lime green": "#65fe08",
    "celadon": "#befdb7",
    "pinkish brown": "#b17261",
    "poo brown": "#885f01",
    "bright sky blue": "#02ccfe",
    "celery": "#c1fd95",
    "dirt brown": "#836539",
    "strawberry": "#fb2943",
    "dark lime": "#84b701",
    "copper": "#b66325",
    "medium brown": "#7f5112",
    "muted green": "#5fa052",
    "robin's egg": "#6dedfd",
    "bright aqua": "#0bf9ea",
    "bright lavender": "#c760ff",
    "ivory": "#ffffcb",
    "very light purple": "#f6cefc",
    "light navy": "#155084",
    "pink red": "#f5054f",
    "olive brown": "#645403",
    "poop brown": "#7a5901",
    "mustard green": "#a8b504",
    "ocean green": "#3d9973",
    "very dark blue": "#000133",
    "dusty green": "#76a973",
    "light navy blue": "#2e5a88",
    "minty green": "#0bf77d",
    "adobe": "#bd6c48",
    "barney": "#ac1db8",
    "jade green": "#2baf6a",
    "bright light blue": "#26f7fd",
    "light lime": "#aefd6c",
    "dark khaki": "#9b8f55",
    "orange yellow": "#ffad01",
    "ocre": "#c69c04",
    "maize": "#f4d054",
    "faded pink": "#de9dac",
    "british racing green": "#05480d",
    "sandstone": "#c9ae74",
    "mud brown": "#60460f",
    "light sea green": "#98f6b0",
    "robin egg blue": "#8af1fe",
    "aqua marine": "#2ee8bb",
    "dark sea green": "#11875d",
    "soft pink": "#fdb0c0",
    "orangey brown": "#b16002",
    "cherry red": "#f7022a",
    "burnt yellow": "#d5ab09",
    "brownish grey": "#86775f",
    "camel": "#c69f59",
    "purplish grey": "#7a687f",
    "marine": "#042e60",
    "greyish pink": "#c88d94",
    "pale turquoise": "#a5fbd5",
    "pastel yellow": "#fffe71",
    "bluey purple": "#6241c7",
    "canary yellow": "#fffe40",
    "faded red": "#d3494e",
    "sepia": "#985e2b",
    "coffee": "#a6814c",
    "bright magenta": "#ff08e8",
    "mocha": "#9d7651",
    "ecru": "#feffca",
    "purpleish": "#98568d",
    "cranberry": "#9e003a",
    "darkish green": "#287c37",
    "brown orange": "#b96902",
    "dusky rose": "#ba6873",
    "melon": "#ff7855",
    "sickly green": "#94b21c",
    "silver": "#c5c9c7",
    "purply blue": "#661aee",
    "purpleish blue": "#6140ef",
    "hospital green": "#9be5aa",
    "shit brown": "#7b5804",
    "mid blue": "#276ab3",
    "amber": "#feb308",
    "easter green": "#8cfd7e",
    "soft blue": "#6488ea",
    "cerulean blue": "#056eee",
    "golden brown": "#b27a01",
    "bright turquoise": "#0ffef9",
    "red pink": "#fa2a55",
    "red purple": "#820747",
    "greyish brown": "#7a6a4f",
    "vermillion": "#f4320c",
    "russet": "#a13905",
    "steel grey": "#6f828a",
    "lighter purple": "#a55af4",
    "bright violet": "#ad0afd",
    "prussian blue": "#004577",
    "slate green": "#658d6d",
    "dirty pink": "#ca7b80",
    "dark blue green": "#005249",
    "pine": "#2b5d34",
    "yellowy green": "#bff128",
    "dark gold": "#b59410",
    "bluish": "#2976bb",
    "darkish blue": "#014182",
    "dull red": "#bb3f3f",
    "pinky red": "#fc2647",
    "bronze": "#a87900",
    "pale teal": "#82cbb2",
    "military green": "#667c3e",
    "barbie pink": "#fe46a5",
    "bubblegum pink": "#fe83cc",
    "pea soup green": "#94a617",
    "dark mustard": "#a88905",
    "shit": "#7f5f00",
    "medium purple": "#9e43a2",
    "very dark green": "#062e03",
    "dirt": "#8a6e45",
    "dusky pink": "#cc7a8b",
    "red violet": "#9e0168",
    "lemon yellow": "#fdff38",
    "pistachio": "#c0fa8b",
    "dull yellow": "#eedc5b",
    "dark lime green": "#7ebd01",
    "denim blue": "#3b5b92",
    "teal blue": "#01889f",
    "lightish blue": "#3d7afd",
    "purpley blue": "#5f34e7",
    "light indigo": "#6d5acf",
    "swamp green": "#748500",
    "brown green": "#706c11",
    "dark maroon": "#3c0008",
    "hot purple": "#cb00f5",
    "dark forest green": "#002d04",
    "faded blue": "#658cbb",
    "drab green": "#749551",
    "light lime green": "#b9ff66",
    "snot green": "#9dc100",
    "yellowish": "#faee66",
    "light blue green": "#7efbb3",
    "bordeaux": "#7b002c",
    "light mauve": "#c292a1",
    "ocean": "#017b92",
    "marigold": "#fcc006",
    "muddy green": "#657432",
    "dull orange": "#d8863b",
    "steel": "#738595",
    "electric purple": "#aa23ff",
    "fluorescent green": "#08ff08",
    "yellowish brown": "#9b7a01",
    "blush": "#f29e8e",
    "soft green": "#6fc276",
    "bright orange": "#ff5b00",
    "lemon": "#fdff52",
    "purple grey": "#866f85",
    "acid green": "#8ffe09",
    "pale lavender": "#eecffe",
    "violet blue": "#510ac9",
    "light forest green": "#4f9153",
    "burnt red": "#9f2305",
    "khaki green": "#728639",
    "cerise": "#de0c62",
    "faded purple": "#916e99",
    "apricot": "#ffb16d",
    "dark olive green": "#3c4d03",
    "grey brown": "#7f7053",
    "green grey": "#77926f",
    "true blue": "#010fcc",
    "pale violet": "#ceaefa",
    "periwinkle blue": "#8f99fb",
    "light sky blue": "#c6fcff",
    "blurple": "#5539cc",
    "green brown": "#544e03",
    "bluegreen": "#017a79",
    "bright teal": "#01f9c6",
    "brownish yellow": "#c9b003",
    "pea soup": "#929901",
    "forest": "#0b5509",
    "barney purple": "#a00498",
    "ultramarine": "#2000b1",
    "purplish": "#94568c",
    "puke yellow": "#c2be0e",
    "bluish grey": "#748b97",
    "dark periwinkle": "#665fd1",
    "dark lilac": "#9c6da5",
    "reddish": "#c44240",
    "light maroon": "#a24857",
    "dusty purple": "#825f87",
    "terra cotta": "#c9643b",
    "avocado": "#90b134",
    "marine blue": "#01386a",
    "teal green": "#25a36f",
    "slate grey": "#59656d",
    "lighter green": "#75fd63",
    "electric green": "#21fc0d",
    "dusty blue": "#5a86ad",
    "golden yellow": "#fec615",
    "bright yellow": "#fffd01",
    "light lavender": "#dfc5fe",
    "umber": "#b26400",
    "poop": "#7f5e00",
    "dark peach": "#de7e5d",
    "jungle green": "#048243",
    "eggshell": "#ffffd4",
    "denim": "#3b638c",
    "yellow brown": "#b79400",
    "dull purple": "#84597e",
    "chocolate brown": "#411900",
    "wine red": "#7b0323",
    "neon blue": "#04d9ff",
    "dirty green": "#667e2c",
    "light tan": "#fbeeac",
    "ice blue": "#d7fffe",
    "cadet blue": "#4e7496",
    "dark mauve": "#874c62",
    "very light blue": "#d5ffff",
    "grey purple": "#826d8c",
    "pastel pink": "#ffbacd",
    "very light green": "#d1ffbd",
    "dark sky blue": "#448ee4",
    "evergreen": "#05472a",
    "dull pink": "#d5869d",
    "aubergine": "#3d0734",
    "mahogany": "#4a0100",
    "reddish orange": "#f8481c",
    "deep green": "#02590f",
    "vomit green": "#89a203",
    "purple pink": "#e03fd8",
    "dusty pink": "#d58a94",
    "faded green": "#7bb274",
    "camo green": "#526525",
    "pinky purple": "#c94cbe",
    "pink purple": "#db4bda",
    "brownish red": "#9e3623",
    "dark rose": "#b5485d",
    "mud": "#735c12",
    "brownish": "#9c6d57",
    "emerald green": "#028f1e",
    "pale brown": "#b1916e",
    "dull blue": "#49759c",
    "burnt umber": "#a0450e",
    "medium green": "#39ad48",
    "clay": "#b66a50",
    "light aqua": "#8cffdb",
    "light olive green": "#a4be5c",
    "brownish orange": "#cb7723",
    "dark aqua": "#05696b",
    "purplish pink": "#ce5dae",
    "dark salmon": "#c85a53",
    "greenish grey": "#96ae8d",
    "jade": "#1fa774",
    "ugly green": "#7a9703",
    "dark beige": "#ac9362",
    "emerald": "#01a049",
    "pale red": "#d9544d",
    "light magenta": "#fa5ff7",
    "sky": "#82cafc",
    "light cyan": "#acfffc",
    "yellow orange": "#fcb001",
    "reddish purple": "#910951",
    "reddish pink": "#fe2c54",
    "orchid": "#c875c4",
    "dirty yellow": "#cdc50a",
    "orange red": "#fd411e",
    "deep red": "#9a0200",
    "orange brown": "#be6400",
    "cobalt blue": "#030aa7",
    "neon pink": "#fe019a",
    "rose pink": "#f7879a",
    "greyish purple": "#887191",
    "raspberry": "#b00149",
    "aqua green": "#12e193",
    "salmon pink": "#fe7b7c",
    "tangerine": "#ff9408",
    "brownish green": "#6a6e09",
    "red brown": "#8b2e16",
    "greenish brown": "#696112",
    "pumpkin": "#e17701",
    "pine green": "#0a481e",
    "charcoal": "#343837",
    "baby pink": "#ffb7ce",
    "cornflower": "#6a79f7",
    "blue violet": "#5d06e9",
    "chocolate": "#3d1c02",
    "greyish green": "#82a67d",
    "scarlet": "#be0119",
    "green yellow": "#c9ff27",
    "dark olive": "#373e02",
    "sienna": "#a9561e",
    "pastel purple": "#caa0ff",
    "terracotta": "#ca6641",
    "aqua blue": "#02d8e9",
    "sage green": "#88b378",
    "blood red": "#980002",
    "deep pink": "#cb0162",
    "grass": "#5cac2d",
    "moss": "#769958",
    "pastel blue": "#a2bffe",
    "bluish green": "#10a674",
    "green blue": "#06b48b",
    "dark tan": "#af884a",
    "greenish blue": "#0b8b87",
    "pale orange": "#ffa756",
    "vomit": "#a2a415",
    "forrest green": "#154406",
    "dark lavender": "#856798",
    "dark violet": "#34013f",
    "purple blue": "#632de9",
    "dark cyan": "#0a888a",
    "olive drab": "#6f7632",
    "pinkish": "#d46a7e",
    "cobalt": "#1e488f",
    "neon purple": "#bc13fe",
    "light turquoise": "#7ef4cc",
    "apple green": "#76cd26",
    "dull green": "#74a662",
    "wine": "#80013f",
    "powder blue": "#b1d1fc",
    "off white": "#ffffe4",
    "electric blue": "#0652ff",
    "dark turquoise": "#045c5a",
    "blue purple": "#5729ce",
    "azure": "#069af3",
    "bright red": "#ff000d",
    "pinkish red": "#f10c45",
    "cornflower blue": "#5170d7",
    "light olive": "#acbf69",
    "grape": "#6c3461",
    "greyish blue": "#5e819d",
    "purplish blue": "#601ef9",
    "yellowish green": "#b0dd16",
    "greenish yellow": "#cdfd02",
    "medium blue": "#2c6fbb",
    "dusty rose": "#c0737a",
    "light violet": "#d6b4fc",
    "midnight blue": "#020035",
    "bluish purple": "#703be7",
    "red orange": "#fd3c06",
    "dark magenta": "#960056",
    "greenish": "#40a368",
    "ocean blue": "#03719c",
    "coral": "#fc5a50",
    "cream": "#ffffc2",
    "reddish brown": "#7f2b0a",
    "burnt sienna": "#b04e0f",
    "brick": "#a03623",
    "sage": "#87ae73",
    "grey green": "#789b73",
    "white": "#ffffff",
    "robin's egg blue": "#98eff9",
    "moss green": "#658b38",
    "steel blue": "#5a7d9a",
    "eggplant": "#380835",
    "light yellow": "#fffe7a",
    "leaf green": "#5ca904",
    "light grey": "#d8dcd6",
    "puke": "#a5a502",
    "pinkish purple": "#d648d7",
    "sea blue": "#047495",
    "pale purple": "#b790d4",
    "slate blue": "#5b7c99",
    "blue grey": "#607c8e",
    "hunter green": "#0b4008",
    "fuchsia": "#ed0dd9",
    "crimson": "#8c000f",
    "pale yellow": "#ffff84",
    "ochre": "#bf9005",
    "mustard yellow": "#d2bd0a",
    "light red": "#ff474c",
    "cerulean": "#0485d1",
    "pale pink": "#ffcfdc",
    "deep blue": "#040273",
    "rust": "#a83c09",
    "light teal": "#90e4c1",
    "slate": "#516572",
    "goldenrod": "#fac205",
    "dark yellow": "#d5b60a",
    "dark grey": "#363737",
    "army green": "#4b5d16",
    "grey blue": "#6b8ba4",
    "seafoam": "#80f9ad",
    "puce": "#a57e52",
    "spring green": "#a9f971",
    "dark orange": "#c65102",
    "sand": "#e2ca76",
    "pastel green": "#b0ff9d",
    "mint": "#9ffeb0",
    "light orange": "#fdaa48",
    "bright pink": "#fe01b1",
    "chartreuse": "#c1f80a",
    "deep purple": "#36013f",
    "dark brown": "#341c02",
    "taupe": "#b9a281",
    "pea green": "#8eab12",
    "puke green": "#9aae07",
    "kelly green": "#02ab2e",
    "seafoam green": "#7af9ab",
    "blue green": "#137e6d",
    "khaki": "#aaa662",
    "burgundy": "#610023",
    "dark teal": "#014d4e",
    "brick red": "#8f1402",
    "royal purple": "#4b006e",
    "plum": "#580f41",
    "mint green": "#8fff9f",
    "gold": "#dbb40c",
    "baby blue": "#a2cffe",
    "yellow green": "#c0fb2d",
    "bright purple": "#be03fd",
    "dark red": "#840000",
    "pale blue": "#d0fefe",
    "grass green": "#3f9b0b",
    "navy": "#01153e",
    "aquamarine": "#04d8b2",
    "burnt orange": "#c04e01",
    "neon green": "#0cff0c",
    "bright blue": "#0165fc",
    "rose": "#cf6275",
    "light pink": "#ffd1df",
    "mustard": "#ceb301",
    "indigo": "#380282",
    "lime": "#aaff32",
    "sea green": "#53fca1",
    "periwinkle": "#8e82fe",
    "dark pink": "#cb416b",
    "olive green": "#677a04",
    "peach": "#ffb07c",
    "pale green": "#c7fdb5",
    "light brown": "#ad8150",
    "hot pink": "#ff028d",
    "black": "#000000",
    "lilac": "#cea2fd",
    "navy blue": "#001146",
    "royal blue": "#0504aa",
    "beige": "#e6daa6",
    "salmon": "#ff796c",
    "olive": "#6e750e",
    "maroon": "#650021",
    "bright green": "#01ff07",
    "dark purple": "#35063e",
    "mauve": "#ae7181",
    "forest green": "#06470c",
    "aqua": "#13eac9",
    "cyan": "#00ffff",
    "tan": "#d1b26f",
    "dark blue": "#00035b",
    "lavender": "#c79fef",
    "turquoise": "#06c2ac",
    "dark green": "#033500",
    "violet": "#9a0eea",
    "light purple": "#bf77f6",
    "lime green": "#89fe05",
    "grey": "#929591",
    "sky blue": "#75bbfd",
    "yellow": "#ffff14",
    "magenta": "#c20078",
    "light green": "#96f97b",
    "orange": "#f97306",
    "teal": "#029386",
    "light blue": "#95d0fc",
    "red": "#e50000",
    "brown": "#653700",
    "pink": "#ff81c0",
    "blue": "#0343df",
    "green": "#15b01a",
    "purple": "#7e1e9c",
}

# Normalize name to "xkcd:<name>" to avoid name collisions.
XKCD_COLORS = {"xkcd:" + name: value for name, value in XKCD_COLORS.items()}


# https://drafts.csswg.org/css-color-4/#named-colors
CSS4_COLORS = {
    "aliceblue": "#F0F8FF",
    "antiquewhite": "#FAEBD7",
    "aqua": "#00FFFF",
    "aquamarine": "#7FFFD4",
    "azure": "#F0FFFF",
    "beige": "#F5F5DC",
    "bisque": "#FFE4C4",
    "black": "#000000",
    "blanchedalmond": "#FFEBCD",
    "blue": "#0000FF",
    "blueviolet": "#8A2BE2",
    "brown": "#A52A2A",
    "burlywood": "#DEB887",
    "cadetblue": "#5F9EA0",
    "chartreuse": "#7FFF00",
    "chocolate": "#D2691E",
    "coral": "#FF7F50",
    "cornflowerblue": "#6495ED",
    "cornsilk": "#FFF8DC",
    "crimson": "#DC143C",
    "cyan": "#00FFFF",
    "darkblue": "#00008B",
    "darkcyan": "#008B8B",
    "darkgoldenrod": "#B8860B",
    "darkgray": "#A9A9A9",
    "darkgreen": "#006400",
    "darkgrey": "#A9A9A9",
    "darkkhaki": "#BDB76B",
    "darkmagenta": "#8B008B",
    "darkolivegreen": "#556B2F",
    "darkorange": "#FF8C00",
    "darkorchid": "#9932CC",
    "darkred": "#8B0000",
    "darksalmon": "#E9967A",
    "darkseagreen": "#8FBC8F",
    "darkslateblue": "#483D8B",
    "darkslategray": "#2F4F4F",
    "darkslategrey": "#2F4F4F",
    "darkturquoise": "#00CED1",
    "darkviolet": "#9400D3",
    "deeppink": "#FF1493",
    "deepskyblue": "#00BFFF",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "dodgerblue": "#1E90FF",
    "firebrick": "#B22222",
    "floralwhite": "#FFFAF0",
    "forestgreen": "#228B22",
    "fuchsia": "#FF00FF",
    "gainsboro": "#DCDCDC",
    "ghostwhite": "#F8F8FF",
    "gold": "#FFD700",
    "goldenrod": "#DAA520",
    "gray": "#808080",
    "green": "#008000",
    "greenyellow": "#ADFF2F",
    "grey": "#808080",
    "honeydew": "#F0FFF0",
    "hotpink": "#FF69B4",
    "indianred": "#CD5C5C",
    "indigo": "#4B0082",
    "ivory": "#FFFFF0",
    "khaki": "#F0E68C",
    "lavender": "#E6E6FA",
    "lavenderblush": "#FFF0F5",
    "lawngreen": "#7CFC00",
    "lemonchiffon": "#FFFACD",
    "lightblue": "#ADD8E6",
    "lightcoral": "#F08080",
    "lightcyan": "#E0FFFF",
    "lightgoldenrodyellow": "#FAFAD2",
    "lightgray": "#D3D3D3",
    "lightgreen": "#90EE90",
    "lightgrey": "#D3D3D3",
    "lightpink": "#FFB6C1",
    "lightsalmon": "#FFA07A",
    "lightseagreen": "#20B2AA",
    "lightskyblue": "#87CEFA",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "lightsteelblue": "#B0C4DE",
    "lightyellow": "#FFFFE0",
    "lime": "#00FF00",
    "limegreen": "#32CD32",
    "linen": "#FAF0E6",
    "magenta": "#FF00FF",
    "maroon": "#800000",
    "mediumaquamarine": "#66CDAA",
    "mediumblue": "#0000CD",
    "mediumorchid": "#BA55D3",
    "mediumpurple": "#9370DB",
    "mediumseagreen": "#3CB371",
    "mediumslateblue": "#7B68EE",
    "mediumspringgreen": "#00FA9A",
    "mediumturquoise": "#48D1CC",
    "mediumvioletred": "#C71585",
    "midnightblue": "#191970",
    "mintcream": "#F5FFFA",
    "mistyrose": "#FFE4E1",
    "moccasin": "#FFE4B5",
    "navajowhite": "#FFDEAD",
    "navy": "#000080",
    "oldlace": "#FDF5E6",
    "olive": "#808000",
    "olivedrab": "#6B8E23",
    "orange": "#FFA500",
    "orangered": "#FF4500",
    "orchid": "#DA70D6",
    "palegoldenrod": "#EEE8AA",
    "palegreen": "#98FB98",
    "paleturquoise": "#AFEEEE",
    "palevioletred": "#DB7093",
    "papayawhip": "#FFEFD5",
    "peachpuff": "#FFDAB9",
    "peru": "#CD853F",
    "pink": "#FFC0CB",
    "plum": "#DDA0DD",
    "powderblue": "#B0E0E6",
    "purple": "#800080",
    "rebeccapurple": "#663399",
    "red": "#FF0000",
    "rosybrown": "#BC8F8F",
    "royalblue": "#4169E1",
    "saddlebrown": "#8B4513",
    "salmon": "#FA8072",
    "sandybrown": "#F4A460",
    "seagreen": "#2E8B57",
    "seashell": "#FFF5EE",
    "sienna": "#A0522D",
    "silver": "#C0C0C0",
    "skyblue": "#87CEEB",
    "slateblue": "#6A5ACD",
    "slategray": "#708090",
    "slategrey": "#708090",
    "snow": "#FFFAFA",
    "springgreen": "#00FF7F",
    "steelblue": "#4682B4",
    "tan": "#D2B48C",
    "teal": "#008080",
    "thistle": "#D8BFD8",
    "tomato": "#FF6347",
    "turquoise": "#40E0D0",
    "violet": "#EE82EE",
    "wheat": "#F5DEB3",
    "white": "#FFFFFF",
    "whitesmoke": "#F5F5F5",
    "yellow": "#FFFF00",
    "yellowgreen": "#9ACD32",
}



from enum import Enum

import pyproj

# Minimum integer representable in a float32
# https://stackoverflow.com/a/3793950
MIN_INTEGER_FLOAT32 = -16777216

# Maximum integer representable in a float32
# https://stackoverflow.com/a/3793950
MAX_INTEGER_FLOAT32 = 16777216

EPSG_4326 = pyproj.CRS("epsg:4326")

# In pyodide, the pyproj PROJ data directory is much smaller, and it currently
# hard-crashes on the line `pyproj.CRS("ogc:84")`. Instead, we vendor the PROJJSON
# representation of this CRS, which works in pyodide.
OGC_84_dict = {
    "$schema": "https://proj.org/schemas/v0.7/projjson.schema.json",
    "type": "GeographicCRS",
    "name": "WGS 84 (CRS84)",
    "datum_ensemble": {
        "name": "World Geodetic System 1984 ensemble",
        "members": [
            {
                "name": "World Geodetic System 1984 (Transit)",
                "id": {"authority": "EPSG", "code": 1166},
            },
            {
                "name": "World Geodetic System 1984 (G730)",
                "id": {"authority": "EPSG", "code": 1152},
            },
            {
                "name": "World Geodetic System 1984 (G873)",
                "id": {"authority": "EPSG", "code": 1153},
            },
            {
                "name": "World Geodetic System 1984 (G1150)",
                "id": {"authority": "EPSG", "code": 1154},
            },
            {
                "name": "World Geodetic System 1984 (G1674)",
                "id": {"authority": "EPSG", "code": 1155},
            },
            {
                "name": "World Geodetic System 1984 (G1762)",
                "id": {"authority": "EPSG", "code": 1156},
            },
            {
                "name": "World Geodetic System 1984 (G2139)",
                "id": {"authority": "EPSG", "code": 1309},
            },
        ],
        "ellipsoid": {
            "name": "WGS 84",
            "semi_major_axis": 6378137,
            "inverse_flattening": 298.257223563,
        },
        "accuracy": "2.0",
        "id": {"authority": "EPSG", "code": 6326},
    },
    "coordinate_system": {
        "subtype": "ellipsoidal",
        "axis": [
            {
                "name": "Geodetic longitude",
                "abbreviation": "Lon",
                "direction": "east",
                "unit": "degree",
            },
            {
                "name": "Geodetic latitude",
                "abbreviation": "Lat",
                "direction": "north",
                "unit": "degree",
            },
        ],
    },
    "scope": "Not known.",
    "area": "World.",
    "bbox": {
        "south_latitude": -90,
        "west_longitude": -180,
        "north_latitude": 90,
        "east_longitude": 180,
    },
    "id": {"authority": "OGC", "code": "CRS84"},
}

OGC_84 = pyproj.CRS.from_json_dict(OGC_84_dict)


class EXTENSION_NAME(bytes, Enum):  # noqa: N801
    """GeoArrow extension name."""

    POINT = b"geoarrow.point"
    LINESTRING = b"geoarrow.linestring"
    POLYGON = b"geoarrow.polygon"
    MULTIPOINT = b"geoarrow.multipoint"
    MULTILINESTRING = b"geoarrow.multilinestring"
    MULTIPOLYGON = b"geoarrow.multipolygon"
    BOX = b"geoarrow.box"
    GEOMETRY = b"geoarrow.geometry"
    GEOMETRYCOLLECTION = b"geoarrow.geometrycollection"
    WKB = b"geoarrow.wkb"
    WKT = b"geoarrow.wkt"
    OGC_WKB = b"ogc.wkb"

    def __str__(self) -> str:
        return self.value.decode()


class Environment(str, Enum):
    Azure = "azure"
    """Azure notebook"""

    Cocalc = "cocalc"
    Colab = "colab"
    """Colab notebook"""

    Databricks = "databricks"

    IPythonTerminal = "ipython_terminal"

    Kaggle = "kaggle"
    """Kaggle notebook"""

    Nteract = "nteract"
    Unknown = "unknown"
    Vscode = "vscode"



from typing import ClassVar

import traitlets as t

from lonboard._base import BaseExtension
from lonboard.traits import (
    DashArrayAccessor,
    FilterValueAccessor,
    FloatAccessor,
    PointAccessor,
    VariableLengthTuple,
)


class BrushingExtension(BaseExtension):
    """Adds GPU-based data brushing functionalities to layers.

    It allows the layer to show/hide objects based on the current pointer position.

    # Example

    An example is in the [County-to-County Migration
    notebook](https://developmentseed.org/lonboard/latest/examples/migration/).

    # Layer Properties

    This extension dynamically enables the following properties onto the layer(s) where
    it is included:

    ## `brushing_enabled`

    Enable/disable brushing. If brushing is disabled, all objects are rendered.

    - Type: `bool`, optional
    - Default: `True`

    ## `brushing_target`

    The position used to filter each object by. One of the following:

    - `"source"`: Use the primary position for each object. This can mean different
      things depending on the layer. It usually refers to the coordinates returned by
      `getPosition` or `getSourcePosition` accessors.
    - `"target"`: Use the secondary position for each object. This may not be available
      in some layers. It usually refers to the coordinates returned by
      `getTargetPosition` accessor.
    - `"source_target"`: Use both the primary position and secondary position for each
      object. Show object if either is in brushing range.
    - `"custom"`: Some layers may not describe their data objects with one or two
      coordinates, for example `PathLayer` and `PolygonLayer`. Use this option with the
      `get_brushing_target` prop to provide a custom position that each object should be
      filtered by.

    - Type: `str`, optional

        One of: "source" | "target" | "source_target" | "custom"

    - Default: `"source"`

    ## `brushing_radius`

    The brushing radius centered at the pointer, in meters. If a data object is within
    this circle, it is rendered; otherwise it is hidden.

    - Type: `float`, optional
    - Default: `10000`

    ## `get_brushing_target`

    An arbitrary position for each object that it will be filtered by.

    Only effective if `brushing_target` is set to `"custom"`.

    - Type: [PointAccessor][lonboard.traits.PointAccessor], optional
        - If a point is provided, it is used as the target for all rows.
        - If an array of points is provided, each value in the array will be used as the
          target for the row at the same row index.
    - Default: `None`.
    """

    _extension_type = t.Unicode("brushing").tag(sync=True)

    _layer_traits: ClassVar = {
        "brushing_enabled": t.Bool(default_value=True).tag(sync=True),
        "brushing_target": t.Unicode(None, allow_none=True).tag(sync=True),
        "brushing_radius": t.Float(None, allow_none=True, min=0).tag(sync=True),
        "get_brushing_target": PointAccessor(None, allow_none=True),
    }


class CollisionFilterExtension(BaseExtension):
    """Allows layers to hide overlapping objects.

    # Layer Properties

    This extension dynamically enables the following properties onto the layer(s) where
    it is included:

    ## `collision_enabled`

    Enable/disable collisions. If collisions are disabled, all objects are rendered.

    - Type: `bool`, optional
    - Default: `True`

    ## `collision_group`

    Collision group this layer belongs to. If it is not set, the 'default' collision
    group is used

    - Type: `str`, optional
    - Default: `None`

    ## `get_collision_priority`

    Accessor for collision priority. Must return a number in the range -1000 -> 1000.
    Features with higher values are shown preferentially.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the priority for all objects.
        - If an array is provided, each value in the array will be used as the priority
          for the object at the same row index.
    - Default: `0`.

    """

    _extension_type = t.Unicode("collision-filter").tag(sync=True)

    _layer_traits: ClassVar = {
        "collision_enabled": t.Bool(default_value=True).tag(sync=True),
        "collision_group": t.Unicode(None, allow_none=True).tag(sync=True),
        "get_collision_priority": FloatAccessor(None, allow_none=True),
    }


class DataFilterExtension(BaseExtension):
    """Adds GPU-based data filtering functionalities to layers.

    It allows the layer to show/hide objects based on user-defined properties.

    # Example

    ```py
    from lonboard import Map, ScatterplotLayer
    from lonboard.colormap import apply_continuous_cmap
    from lonboard.layer_extension import DataFilterExtension

    gdf = gpd.GeoDataFrame(...)
    extension = DataFilterExtension()
    layer = ScatterplotLayer.from_geopandas(
        gdf,
        extensions=[extension],
        get_filter_value=gdf["filter_value"], # replace with desired column
        filter_range=[0, 5] # replace with desired filter range
    )
    ```

    The `DataFilterExtension` allows filtering on 1 to 4 attributes at the same time.
    So if you have four numeric columns of interest, you can filter on the
    intersection of all of them.

    For easy visualization, we suggest connecting the `DataFilterExtension` to an
    interactive slider from `ipywidgets`.

    ```py
    from ipywidgets import FloatRangeSlider

    slider = FloatRangeSlider(
        value=(2, 5),
        min=0,
        max=10,
        step=0.1,
        description="Slider: "
    )
    slider

    jsdlink(
        (slider, "value"),
        (layer, "filter_range")
    )
    ```

    If you have 2 to 4 columns, use a
    [`MultiRangeSlider`][lonboard.controls.MultiRangeSlider], which combines multiple
    `FloatRangeSlider` objects in a form that the `DataFilterExtension` expects.

    ```py
    from ipywidgets import FloatRangeSlider, jsdlink

    slider1 = FloatRangeSlider(
        value=(2, 5),
        min=0,
        max=10,
        step=0.1,
        description="First slider: "
    )
    slider2 = FloatRangeSlider(
        value=(30, 40),
        min=0,
        max=50,
        step=1,
        description="Second slider: "
    )
    multi_slider = MultiRangeSlider([slider1, slider2])
    multi_slider

    jsdlink(
        (multi_slider, "value"),
        (layer, "filter_range")
    )
    ```

    # Important notes

    - The DataFilterExtension only supports float32 data, so integer data will be casted
      to float32.
    - The DataFilterExtension copies all data referenced by `get_filter_value` to the
      GPU, so it will increase memory pressure on the GPU.

    # Layer Properties

    This extension dynamically enables the following properties onto the layer(s) where
    it is included:

    ## `filter_categories`

    The list of categories that should be rendered. If an object's filtered category is
    in the list, the object will be rendered; otherwise it will be hidden. This prop can
    be updated on user input or animation with very little cost.

    Format:

    - If category_size is 1: ['category1', 'category2']
    - If category_size is 2 to 4:
        [['category1', 'category2', ...], ['category3', ...], ...] for each filtered
        property, respectively.

    The maximum number of supported is determined by the category_size:

    - If category_size is 1: 128 categories
    - If category_size is 2: 64 categories per dimension
    - If category_size is 3 or 4: 32 categories per dimension.

    If this value is exceeded any categories beyond the limit will be ignored.

    Default: `[0]`

    ## `filter_enabled`

    Enable/disable the data filter. If the data filter is disabled, all objects are
    rendered.

    - Type: `bool`, optional
    - Default: `True`

    ## `filter_range`

    The bounds which defines whether an object should be rendered. If an object's
    filtered value is within the bounds, the object will be rendered; otherwise it will
    be hidden. This prop can be updated on user input or animation with very little
    cost.

    Format:

    If `filter_size` is 1, provide a single tuple of `(min, max)`.

    If `filter_size` is 2 to 4, provide a list of tuples: `[(min0, max0), (min1,
    max1), ...]` for each filtered property, respectively.

    - Type: either Tuple[float, float] or List[Tuple[float, float]], optional
    - Default: `(-1, 1)`

    ## `filter_soft_range`

    If specified, objects will be faded in/out instead of abruptly shown/hidden.

    When the filtered value is outside of the bounds defined by `filter_soft_range` but
    still within the bounds defined by `filter_range`, the object will be rendered as
    "faded".

    - Type: Tuple[float, float], optional
    - Default: `None`

    ## `filter_transform_size`

    When an object is "faded", manipulate its size so that it appears smaller or
    thinner. Only works if `filter_soft_range` is specified.

    - Type: `bool`, optional
    - Default: `True`

    ## `filter_transform_color`

    When an object is "faded", manipulate its opacity so that it appears more
    translucent. Only works if `filter_soft_range` is specified.

    - Type: `bool`, optional
    - Default: `True`

    ## `get_filter_value`

    Accessor to retrieve the value for each object that it will be filtered by.

    - Type: [FilterValueAccessor][lonboard.traits.FilterValueAccessor]
        - If a scalar value is provided, it is used as the value for all objects.
        - If an array is provided, each value in the array will be used as the value for
          the object at the same row index.

    ## `get_filter_category`

    Accessor to retrieve the category for each object that it will be filtered by.

    - Type: [FilterValueAccessor][lonboard.traits.FilterValueAccessor]
        - If a scalar value is provided, it is used as the value for all objects.
        - If an array is provided, each value in the array will be used as the value for
          the object at the same row index.
    """

    _extension_type = t.Unicode("data-filter").tag(sync=True)

    _layer_traits: ClassVar = {
        "filter_categories": t.Union(
            [
                VariableLengthTuple(t.Any()),
                VariableLengthTuple(
                    VariableLengthTuple(t.Any()),
                    minlen=2,
                    maxlen=4,
                ),
            ],
            default_value=None,
            allow_none=True,
        ).tag(sync=True),
        "filter_enabled": t.Bool(default_value=True).tag(sync=True),
        "filter_range": t.Union(
            [
                VariableLengthTuple(t.Float(), minlen=2, maxlen=2),
                VariableLengthTuple(
                    VariableLengthTuple(t.Float(), minlen=2, maxlen=2),
                    minlen=2,
                    maxlen=4,
                ),
            ],
            default_value=None,
            allow_none=True,
        ).tag(sync=True),
        "filter_soft_range": t.Tuple(
            t.Float(),
            t.Float(),
            default_value=None,
            allow_none=True,
        ).tag(sync=True),
        "filter_transform_size": t.Bool(default_value=True).tag(sync=True),
        "filter_transform_color": t.Bool(default_value=True).tag(sync=True),
        "get_filter_value": FilterValueAccessor(default_value=None, allow_none=True),
        "get_filter_category": FilterValueAccessor(default_value=None, allow_none=True),
    }

    filter_size = t.Int(None, min=1, max=4, allow_none=True).tag(sync=True)
    """The size of the filter (number of columns to filter by).

    The data filter can show/hide data based on 1-4 numeric properties of each object.

    - Type: `int`. This is required if using range-based filtering.
    - Default 1.
    """

    category_size = t.Int(None, min=1, max=4, allow_none=True).tag(sync=True)
    """The size of the category filter (number of columns to filter by).

    The category filter can show/hide data based on 1-4 properties of each object.

    - Type: `int`. This is required if using category-based filtering.
    - Default 0.
    """


class PathStyleExtension(BaseExtension):
    """Adds selected features to the PathLayer and composite layers that render
    the [PathLayer][lonboard.PathLayer], e.g. [PolygonLayer][lonboard.PolygonLayer].

    # Example

    Deck.gl example https://deck.gl/docs/api-reference/extensions/path-style-extension

    # Layer Properties

    This extension dynamically enables the following properties onto the layer(s) where
    it is included:

    ## `get_dash_array`

    Must be specified if the dash option is enabled. The dash array to draw each path
    with `[dash_size, gap_size]` relative to the width of the path.

    - Type: [DashArrayAccessor][lonboard.traits.DashArrayAccessor], optional
        - If a scalar value is provided, it is used as the value for all objects. Pass
          `[0, 0]` to draw the path in solid line.
        - If an array is provided, each value in the array will be used as the value
          for the object at the same row index.
    - Default: `[0, 0]`.

    ## `dash_justified`

    Only effective if `get_dash_array` is specified. If `True`, adjust gaps for the
    dashes to align at both ends. Overrides the effect of `high_precision_dash`.

    - Type: `boolean`, optional
    - Default `False`.

    ## `get_offset`

    Must be specified if the `offset` option is enabled.

    The offset to draw each path with, relative to the width of the path. Negative
    offset is to the left hand side, and positive offset is to the right hand side. `0`
    extrudes the path so that it is centered at the specified coordinates.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the offset for all objects.
        - If an array is provided, each value in the array will be used as the offset
          for the object at the same row index.
    - Default: `None`.

    ## `dash_gap_pickable`

    Only effective if `get_dash_array` is specified. If `True`, gaps between solid
    strokes are pickable. If `False`, only the solid strokes are pickable.

    # Remarks

    ## Limitations

    WebGL2 has guaranteed support for up to 16 attributes per shader. The current
    implementation of `PathLayer` uses 13 attributes. Each one of the options of this
    extension adds one more attribute. In other words, if all options are enabled, the
    layer will not be able to use other extensions.

    ## Tips on Rendering Dash Lines

    There are three modes to render dash lines with this extension:

    1. Default: dash starts from the beginning of each line segment
    2. Justified: dash is stretched to center on each line segment
    3. High precision: dash is evaluated continuously from the beginning of a path

    ![Comparison between dash modes](https://user-images.githubusercontent.com/2059298/93418881-33555280-f860-11ea-82cc-b57ecf2e48ce.png)

    The above table illustrates the visual behavior of the three modes.

    The default mode works best if the data consists of long, disjoint paths. It renders
    dashes at exactly the defined lengths.

    The justified mode is guaranteed to render sharp, well-defined corners. This is
    great for rendering polyline shapes. However, the gap size may look inconsistent
    across line segments due to stretching.

    The high precision mode pre-calculates path length on the CPU, so it may be slower
    and use more resources for large datasets. When a path contains a lot of short
    segments, this mode yields the best result.
    """  # noqa: D205

    _extension_type = t.Unicode("path-style").tag(sync=True)

    _layer_traits: ClassVar = {
        "get_dash_array": DashArrayAccessor(None, allow_none=True),
        "dash_justified": t.Bool(None, allow_none=True).tag(sync=True),
        "get_offset": FloatAccessor(None, allow_none=True),
        "dash_gap_pickable": t.Bool(None, allow_none=True).tag(sync=True),
    }

    dash = t.Bool(None, allow_none=True).tag(sync=True)
    """Add capability to render dashed lines.

    - Type: `boolean`, optional
    - Default False.
    """

    high_precision_dash = t.Bool(None, allow_none=True).tag(sync=True)
    """Improve dash rendering quality in certain circumstances.

    Note that this option introduces additional performance overhead. See "Remarks".

    - Type: `boolean`, optional
    - Default `False`.
    """

    offset = t.Bool(None, allow_none=True).tag(sync=True)
    """Add capability to offset lines.

    - Type: `boolean`, optional
    - Default `False`.
    """



"""Compute the weighted centroid of geometries."""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from arro3.core import Array, ChunkedArray, DataType, Field, list_flatten, struct_field

from lonboard._constants import EXTENSION_NAME


@dataclass
class WeightedCentroid:
    # Existing average for x and y
    x: float | None = None
    y: float | None = None
    num_items: int = 0

    def update(self, other: WeightedCentroid) -> None:
        new_chunk_len = other.num_items

        if self.x is None or self.y is None:
            assert self.x is None and self.y is None and self.num_items == 0
            self.x = other.x
            self.y = other.y
            self.num_items = new_chunk_len
            return

        if other.x is None or other.y is None or other.num_items == 0:
            # Can't update from an uninitialized centroid
            return

        existing_modifier = self.num_items / (self.num_items + new_chunk_len)
        new_chunk_modifier = new_chunk_len / (self.num_items + new_chunk_len)

        new_chunk_avg_x = other.x
        new_chunk_avg_y = other.y

        existing_x_avg = self.x
        existing_y_avg = self.y

        self.x = (
            existing_x_avg * existing_modifier + new_chunk_avg_x * new_chunk_modifier
        )
        self.y = (
            existing_y_avg * existing_modifier + new_chunk_avg_y * new_chunk_modifier
        )
        self.num_items += new_chunk_len

    def update_coords(self, coords: Array) -> None:
        """Update the average for x and y based on a new chunk of data.

        Note that this does not keep a cumulative sum due to precision concerns. Rather
        it incrementally updates based on a delta, and never multiplies to large
        constant values.

        Note: this currently computes the mean weighted _per coordinate_ and not _per
        geometry_.
        """
        assert DataType.is_fixed_size_list(coords.type)
        list_size = coords.type.list_size
        assert list_size is not None

        np_arr = list_flatten(coords).to_numpy().reshape(-1, list_size)
        new_chunk_len = np_arr.shape[0]

        if self.x is None or self.y is None:
            assert self.x is None and self.y is None and self.num_items == 0
            self.x = float(np.mean(np_arr[:, 0]))
            self.y = float(np.mean(np_arr[:, 1]))
            self.num_items = new_chunk_len
            return

        existing_modifier = self.num_items / (self.num_items + new_chunk_len)
        new_chunk_modifier = new_chunk_len / (self.num_items + new_chunk_len)

        new_chunk_avg_x = np.mean(np_arr[:, 0])
        new_chunk_avg_y = np.mean(np_arr[:, 1])

        existing_x_avg = self.x
        existing_y_avg = self.y

        self.x = float(
            existing_x_avg * existing_modifier + new_chunk_avg_x * new_chunk_modifier,
        )
        self.y = float(
            existing_y_avg * existing_modifier + new_chunk_avg_y * new_chunk_modifier,
        )
        self.num_items += new_chunk_len


def weighted_centroid(field: Field, column: ChunkedArray) -> WeightedCentroid:
    """Get the bounding box and geometric (weighted) center.

    Of all geometries in the table.
    """
    extension_type_name = field.metadata[b"ARROW:extension:name"]

    if extension_type_name == EXTENSION_NAME.POINT:
        return _weighted_centroid_nest_0(column)

    if extension_type_name in [EXTENSION_NAME.LINESTRING, EXTENSION_NAME.MULTIPOINT]:
        return _weighted_centroid_nest_1(column)

    if extension_type_name in [EXTENSION_NAME.POLYGON, EXTENSION_NAME.MULTILINESTRING]:
        return _weighted_centroid_nest_2(column)

    if extension_type_name == EXTENSION_NAME.MULTIPOLYGON:
        return _weighted_centroid_nest_3(column)

    if extension_type_name == EXTENSION_NAME.BOX:
        return _weighted_centroid_box(column)

    assert False


def _weighted_centroid_nest_0(column: ChunkedArray) -> WeightedCentroid:
    centroid = WeightedCentroid()
    for chunk in column.chunks:
        coords = chunk
        centroid.update_coords(coords)

    return centroid


def _weighted_centroid_nest_1(column: ChunkedArray) -> WeightedCentroid:
    centroid = WeightedCentroid()
    flat_array = list_flatten(column)
    for coords in flat_array:
        centroid.update_coords(coords)

    return centroid


def _weighted_centroid_nest_2(column: ChunkedArray) -> WeightedCentroid:
    centroid = WeightedCentroid()
    flat_array = list_flatten(list_flatten(column))
    for coords in flat_array:
        centroid.update_coords(coords)

    return centroid


def _weighted_centroid_nest_3(column: ChunkedArray) -> WeightedCentroid:
    centroid = WeightedCentroid()
    flat_array = list_flatten(list_flatten(list_flatten(column)))
    for coords in flat_array:
        centroid.update_coords(coords)

    return centroid


def _weighted_centroid_box(column: ChunkedArray) -> WeightedCentroid:
    """Compute the weighted centroid of a box geometry."""
    centroid = WeightedCentroid()
    for chunk in column.chunks:
        is_2d = len(chunk.field.type.fields) == 4
        is_3d = len(chunk.field.type.fields) == 6

        if is_2d:
            minx = struct_field(chunk, 0)
            miny = struct_field(chunk, 1)
            maxx = struct_field(chunk, 2)
            maxy = struct_field(chunk, 3)
        elif is_3d:
            minx = struct_field(chunk, 0)
            miny = struct_field(chunk, 1)
            maxx = struct_field(chunk, 3)
            maxy = struct_field(chunk, 4)
        else:
            raise ValueError(
                f"Unexpected box type with {len(chunk.field.type.fields)} fields.\n"
                "Only 2D and 3D boxes are supported.",
            )

        meanx = float((np.mean(minx) + np.mean(maxx)) / 2)
        meany = float((np.mean(miny) + np.mean(maxy)) / 2)

        centroid.update(WeightedCentroid(x=meanx, y=meany, num_items=len(chunk)))

    return centroid



# Notes:
# - When we pass a value of `None` as a default value to a trait, that value will be
#   serialized to JS as `null` and will not be passed into the GeoArrow model (see the
#   lengthy assignments of type `..(isDefined(this.param) && { param: this.param })`).
#   Then the default value in the JS GeoArrow layer (defined in
#   `@geoarrow/deck.gl-layers`) will be used.

# ruff: noqa: D205, ERA001, SLF001
# 1 blank line required between summary line and description
# Found commented-out code
# Private member accessed: `_layer_traits`


from __future__ import annotations

import warnings
from textwrap import dedent
from typing import TYPE_CHECKING, Any

import ipywidgets
import traitlets
import traitlets as t
from arro3.core import ChunkedArray, Schema, Table

from lonboard._base import BaseExtension, BaseWidget
from lonboard._constants import EXTENSION_NAME, OGC_84
from lonboard._geoarrow._duckdb import from_duckdb as _from_duckdb
from lonboard._geoarrow.c_stream_import import import_arrow_c_stream
from lonboard._geoarrow.geopandas_interop import geopandas_to_geoarrow
from lonboard._geoarrow.ops import reproject_table
from lonboard._geoarrow.ops.bbox import Bbox, total_bounds
from lonboard._geoarrow.ops.centroid import WeightedCentroid, weighted_centroid
from lonboard._geoarrow.ops.coord_layout import make_geometry_interleaved
from lonboard._geoarrow.parse_wkb import parse_serialized_table
from lonboard._geoarrow.row_index import add_positional_row_index
from lonboard._serialization import infer_rows_per_chunk
from lonboard._utils import auto_downcast as _auto_downcast
from lonboard._utils import get_geometry_column_index, remove_extension_kwargs
from lonboard.traits import (
    ArrowTableTrait,
    ColorAccessor,
    FloatAccessor,
    NormalAccessor,
    VariableLengthTuple,
)

if TYPE_CHECKING:
    import sys
    from collections.abc import Sequence

    import geopandas as gpd
    from arro3.core.types import ArrowStreamExportable

    from lonboard.types.layer import (
        BaseLayerKwargs,
        BitmapLayerKwargs,
        BitmapTileLayerKwargs,
        ColumnLayerKwargs,
        HeatmapLayerKwargs,
        PathLayerKwargs,
        PointCloudLayerKwargs,
        PolygonLayerKwargs,
        ScatterplotLayerKwargs,
        SolidPolygonLayerKwargs,
    )

if TYPE_CHECKING:
    import duckdb
    import pyproj

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


class BaseLayer(BaseWidget):
    # Note: these class attributes are **not** serialized to JS
    _bbox = Bbox()
    _weighted_centroid = WeightedCentroid()

    # The following traitlets **are** serialized to JS

    def __init__(
        self,
        *,
        extensions: Sequence[BaseExtension] = (),
        **kwargs: Any,
    ) -> None:
        # We allow layer extensions to dynamically inject properties onto the layer
        # widgets where the layer is defined. We wish to allow extensions and their
        # properties to be passed in the layer constructor. _However_, if

        extension_kwargs = remove_extension_kwargs(extensions, kwargs)

        super().__init__(extensions=extensions, **kwargs)

        # Dynamically set layer traits from extensions after calling __init__
        self._add_extension_traits(extensions)

        # Assign any extension properties that we took out before calling __init__
        added_names: list[str] = []
        for prop_name, prop_value in extension_kwargs.items():
            self.set_trait(prop_name, prop_value)
            added_names.append(prop_name)

        self.send_state(added_names)

    # TODO: validate that only one extension per type is included. E.g. you can't have
    # two data filter extensions.
    extensions = VariableLengthTuple(t.Instance(BaseExtension)).tag(
        sync=True,
        **ipywidgets.widget_serialization,
    )
    """
    A list of [layer extension](https://developmentseed.org/lonboard/latest/api/layer-extensions/)
    objects to add additional features to a layer.
    """

    def _add_extension_traits(self, extensions: Sequence[BaseExtension]) -> None:
        """Assign selected traits from the extension onto this Layer."""
        for extension in extensions:
            # NOTE: here it's important that we call `traitlets.HasTraits.add_traits`
            # and not `self.add_traits`. This is because `add_traits` is originally
            # defined on `HasTraits` but `ipywidgets.Widget` overrides that method to
            # additionally call `send_state` for any trait that has `"sync"` tagged in
            # its metadata. But this is incompatible with traits that don't have default
            # values.
            #
            # For example, with the DataFilterExtension, we want to dynamically add the
            # `get_filter_value` trait, but require that the user pass in a value. With
            # the `Widget` implementation, `send_state` will fail, even if the user
            # passes in a value, because `send_state` is called before we call
            # `super().__init__()`
            traitlets.HasTraits.add_traits(self, **extension._layer_traits)

            # Note: This is part of `Widget.add_traits` (in the direct superclass) that
            # we skip by calling `traitlets.HasTraits.add_traits`
            for name, trait in extension._layer_traits.items():
                if trait.get_metadata("sync"):
                    self.keys.append(name)

    # This doesn't currently work due to I think some race conditions around syncing
    # traits vs the other parameters.

    # def add_extension(self, extension: BaseExtension, **props):
    #     """Add a new layer extension to an existing layer instance.

    #     Any properties for the added extension should also be passed as keyword
    #     arguments to this function.

    #     Examples:

    #     ```py
    #     from lonboard import ScatterplotLayer
    #     from lonboard.layer_extension import DataFilterExtension

    #     gdf = geopandas.GeoDataFrame(...)
    #     layer = ScatterplotLayer.from_geopandas(gdf)

    #     extension = DataFilterExtension(filter_size=1)
    #     filter_values = gdf["filter_column"]

    #     layer.add_extension(
    #         extension,
    #         get_filter_value=filter_values,
    #         filter_range=[0, 1]
    #     )
    #     ```

    #     Args:
    #         extension: The new extension to add.

    #     Raises:
    #         ValueError: if another extension of the same type already exists on the
    #             layer.
    #     """
    #     if any(isinstance(extension, type(ext)) for ext in self.extensions):
    #         raise ValueError("Only one extension of each type permitted")

    #     with self.hold_trait_notifications():
    #         self._add_extension_traits([extension])
    #         self.extensions += (extension,)

    #         # Assign any extension properties
    #         added_names: List[str] = []
    #         for prop_name, prop_value in props.items():
    #             self.set_trait(prop_name, prop_value)
    #             added_names.append(prop_name)

    #     self.send_state(added_names + ["extensions"])

    pickable = t.Bool(default_value=True).tag(sync=True)
    """
    Whether the layer responds to mouse pointer picking events.

    This must be set to `True` for tooltips and other interactive elements to be
    available. This can also be used to only allow picking on specific layers within a
    map instance.

    Note that picking has some performance overhead in rendering. To get the absolute
    best rendering performance with large data (at the cost of removing interactivity),
    set this to `False`.

    - Type: `bool`
    - Default: `True`
    """

    visible = t.Bool(default_value=True).tag(sync=True)
    """
    Whether the layer is visible.

    Under most circumstances, using the `visible` attribute to control the visibility of
    layers is recommended over removing/adding the layer from the `Map.layers` list.

    In particular, toggling the `visible` attribute will persist the layer on the
    JavaScript side, while removing/adding the layer from the `Map.layers` list will
    re-download and re-render from scratch.

    - Type: `bool`
    - Default: `True`
    """

    opacity = t.Float(1, min=0, max=1).tag(sync=True)
    """
    The opacity of the layer.

    - Type: `float`. Must range between 0 and 1.
    - Default: `1`
    """

    auto_highlight = t.Bool(default_value=False).tag(sync=True)
    """
    When true, the current object pointed to by the mouse pointer (when hovered over) is
    highlighted with `highlightColor`.

    Requires `pickable` to be `True`.

    - Type: `bool`
    - Default: `False`
    """

    highlight_color = VariableLengthTuple(
        t.Int(),
        default_value=None,
        minlen=3,
        maxlen=4,
    )
    """
    RGBA color to blend with the highlighted object (the hovered over object if
    `auto_highlight=true`). When the value is a 3 component (RGB) array, a default alpha
    of 255 is applied.

    - Type: List or Tuple of integers
    - Default: `[0, 0, 128, 128]`
    """

    selected_index = t.Int(None, allow_none=True).tag(sync=True)
    """
    The positional index of the most-recently clicked on row of data.

    You can use this to access the full row of data from a GeoDataFrame

    ```py
    gdf.iloc[layer.selected_index]
    ```

    Setting a value here from Python will do nothing. This attribute only exists to be
    updated from JavaScript on a map click. Note that `pickable` must be `True` (the
    default) on this layer for the JavaScript `onClick` handler to work; if `pickable`
    is set to `False`, `selected_index` will never update.

    Note that you can use `observe` to call a function whenever a new value is received
    from JavaScript. Refer
    [here](https://ipywidgets.readthedocs.io/en/stable/examples/Widget%20Events.html#signatures)
    for an example.
    """


def default_geoarrow_viewport(
    table: Table,
) -> tuple[Bbox, WeightedCentroid] | None:
    # Note: in the ArcLayer we won't necessarily have a column with a geoarrow
    # extension type/metadata
    geom_col_idx = get_geometry_column_index(table.schema)
    if geom_col_idx is None:
        return None

    geom_field = table.schema.field(geom_col_idx)
    geom_col = table.column(geom_col_idx)

    table_bbox = total_bounds(geom_field, geom_col)
    table_centroid = weighted_centroid(geom_field, geom_col)

    # Check each layer's data _individually_ to ensure that no layer is outside of
    # epsg:4326 bounds
    if table_centroid.num_items > 0:
        if table_centroid.x is not None and (
            table_centroid.x < -180 or table_centroid.x > 180
        ):
            msg = "Longitude of data's center is outside of WGS84 bounds.\n"
            msg += "Is data in WGS84 projection?"
            raise ValueError(msg)

        if table_centroid.y is not None and (
            table_centroid.y < -90 or table_centroid.y > 90
        ):
            msg = "Latitude of data's center is outside of WGS84 bounds.\n"
            msg += "Is data in WGS84 projection?"
            raise ValueError(msg)

    return table_bbox, table_centroid


class BaseArrowLayer(BaseLayer):
    """Any Arrow-based layer should subclass from BaseArrowLayer."""

    # Note: these class attributes are **not** serialized to JS

    # Number of rows per chunk for serializing table and accessor columns.
    #
    # This is a _layer-level_ construct because we need to ensure the main table and all
    # accessors have exactly the same chunking, because each chunk is rendered
    # independently as a separate deck.gl layer
    _rows_per_chunk: int

    # The following traitlets **are** serialized to JS

    table: ArrowTableTrait

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[BaseLayerKwargs],
    ) -> None:
        """Construct a Layer from a [GeoArrow] table.

        [GeoArrow]: https://geoarrow.org/

        This accepts Arrow data from any library implementing the [Arrow PyCapsule
        Interface], including pyarrow, arro3, DuckDB, and others.

        [Arrow PyCapsule Interface]: https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html

        The geometry column will be reprojected to `EPSG:4326` if it is not already in
        that coordinate system.

        Args:
            table: An Arrow table or stream object from a library implementing the [Arrow PyCapsule Interface]. This object must contain a column with a geometry type that has the `geoarrow` extension metadata.

        Keyword Args:
            kwargs: parameters passed on to `__init__`

        Returns:
            A Layer with the initialized data.

        """
        imported_stream = import_arrow_c_stream(table)
        if isinstance(imported_stream, Table):
            table_o3 = imported_stream
        else:
            assert isinstance(imported_stream, ChunkedArray)
            schema = Schema([imported_stream.field.with_name("geometry")])
            table_o3 = Table.from_arrays([imported_stream], schema=schema)
            table_o3 = add_positional_row_index(table_o3)

        parsed_tables = parse_serialized_table(table_o3)
        assert len(parsed_tables) == 1, (
            "Mixed geometry type input not supported here. Use the top "
            "level viz() function or separate your geometry types in advance."
        )
        table_o3 = parsed_tables[0]
        table_o3 = make_geometry_interleaved(table_o3)

        # Reproject table to WGS84 if needed
        # Note this must happen before calculating the default viewport
        table_o3 = reproject_table(table_o3, to_crs=OGC_84)

        default_viewport = default_geoarrow_viewport(table_o3)
        if default_viewport is not None:
            self._bbox = default_viewport[0]
            self._weighted_centroid = default_viewport[1]

        rows_per_chunk = _rows_per_chunk or infer_rows_per_chunk(table_o3)
        if rows_per_chunk <= 0:
            raise ValueError("Cannot serialize table with 0 rows per chunk.")

        self._rows_per_chunk = rows_per_chunk

        table_o3 = table_o3.rechunk(max_chunksize=rows_per_chunk)

        super().__init__(table=table_o3, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[BaseLayerKwargs],
    ) -> Self:
        """Construct a Layer from a geopandas GeoDataFrame.

        The GeoDataFrame will be reprojected to `EPSG:4326` if it is not already in that
        coordinate system.

        Args:
            gdf: The GeoDataFrame to set on the layer.

        Keyword Args:
            auto_downcast: If `True`, automatically downcast to smaller-size data types
                if possible without loss of precision. This calls
                [pandas.DataFrame.convert_dtypes][pandas.DataFrame.convert_dtypes] and
                [pandas.to_numeric][pandas.to_numeric] under the hood.
            kwargs: parameters passed on to `__init__`

        Returns:
            A Layer with the initialized data.

        """
        if auto_downcast:
            # Note: we don't deep copy because we don't need to clone geometries
            gdf = _auto_downcast(gdf.copy())  # type: ignore

        table = geopandas_to_geoarrow(gdf)
        return cls(table=table, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[BaseLayerKwargs],
    ) -> Self:
        """Construct a Layer from a duckdb-spatial query.

        DuckDB Spatial does not currently expose coordinate reference system
        information, so **the user must ensure that data has been reprojected to
        EPSG:4326** or pass in the existing CRS of the data in the `crs` keyword
        parameter.

        Args:
            sql: The SQL input to visualize. This can either be a string containing a
                SQL query or the output of the duckdb `sql` function.
            con: The current DuckDB connection. This is required when passing a `str` to
                the `sql` parameter.

        Keyword Args:
            crs: The CRS of the input data. This can either be a string passed to
                `pyproj.CRS.from_user_input` or a `pyproj.CRS` object. Defaults to None.
            kwargs: parameters passed on to `__init__`

        Returns:
            A Layer with the initialized data.

        """
        if isinstance(sql, str):
            assert con is not None, "con must be provided when sql is a str"

            rel = con.sql(sql)
            table = _from_duckdb(rel, crs=crs)
        else:
            table = _from_duckdb(sql, crs=crs)

        return cls(table=table, **kwargs)


class BitmapLayer(BaseLayer):
    """The `BitmapLayer` renders a bitmap (e.g. PNG, JPEG, or WebP) at specified
    boundaries.

    **Example:**

    ```py
    from lonboard import Map, BitmapLayer

    layer = BitmapLayer(
        image='https://raw.githubusercontent.com/visgl/deck.gl-data/master/website/sf-districts.png',
        bounds=[-122.5190, 37.7045, -122.355, 37.829]
    )
    m = Map(layer)
    m
    ```
    """

    def __init__(self, **kwargs: BitmapLayerKwargs) -> None:
        super().__init__(**kwargs)  # type: ignore

    _layer_type = t.Unicode("bitmap").tag(sync=True)

    image = t.Unicode().tag(sync=True)
    """The URL to an image to display.

    - Type: `str`
    """

    bounds = t.Union(
        [
            VariableLengthTuple(t.Float(), minlen=4, maxlen=4),
            VariableLengthTuple(
                VariableLengthTuple(t.Float(), minlen=2, maxlen=2),
                minlen=4,
                maxlen=4,
            ),
        ],
    ).tag(sync=True)
    """The bounds of the image.

    Supported formats:

        - Coordinates of the bounding box of the bitmap `[left, bottom, right, top]`
        - Coordinates of four corners of the bitmap, should follow the sequence of
          `[[left, bottom], [left, top], [right, top], [right, bottom]]`.
    """

    desaturate = t.Float(0, min=0, max=1).tag(sync=True)
    """The desaturation of the bitmap. Between `[0, 1]`.

    - Type: `float`, optional
    - Default: `0`
    """

    transparent_color = VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to use for transparent pixels, in `[r, g, b, a]`.

    - Type: `List[float]`, optional
    - Default: `[0, 0, 0, 0]`
    """

    tint_color = VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to tint the bitmap by, in `[r, g, b]`.

    - Type: `List[float]`, optional
    - Default: `[255, 255, 255]`
    """

    @property
    def _bbox(self) -> Bbox:
        a, b, c, d = self.bounds

        # Four corners
        if isinstance(a, tuple):
            bbox = Bbox()
            bbox.update(Bbox(a[0], a[1], a[0], a[1]))
            bbox.update(Bbox(b[0], b[1], b[0], b[1]))
            bbox.update(Bbox(c[0], c[1], c[0], c[1]))
            bbox.update(Bbox(d[0], d[1], d[0], d[1]))
            return bbox

        return Bbox(a, b, c, d)

    @property
    def _weighted_centroid(self) -> WeightedCentroid:
        bbox = self._bbox
        center_x = (bbox.minx + bbox.maxx) / 2
        center_y = (bbox.miny + bbox.maxy) / 2
        # no idea what weight to put on this; we don't know how many "objects" this
        # image should represent.
        return WeightedCentroid(x=center_x, y=center_y, num_items=100)


class BitmapTileLayer(BaseLayer):
    """The BitmapTileLayer renders image tiles (e.g. PNG, JPEG, or WebP) in the web
    mercator tiling system. Only the tiles visible in the current viewport are loaded
    and rendered.

    **Example:**

    ```py
    from lonboard import Map, BitmapTileLayer

    # We set `max_requests < 0` because `tile.openstreetmap.org` supports HTTP/2.
    layer = BitmapTileLayer(
        data="https://tile.openstreetmap.org/{z}/{x}/{y}.png",
        tile_size=256,
        max_requests=-1,
        min_zoom=0,
        max_zoom=19,
    )
    m = Map(layer)
    ```
    """

    def __init__(self, **kwargs: BitmapTileLayerKwargs) -> None:
        super().__init__(**kwargs)  # type: ignore

    _layer_type = t.Unicode("bitmap-tile").tag(sync=True)

    data = t.Union([t.Unicode(), VariableLengthTuple(t.Unicode(), minlen=1)]).tag(
        sync=True,
    )
    """
    Either a URL template or an array of URL templates from which the tile data should
    be loaded.

    If the value is a string: a URL template. Substrings {x} {y} and {z}, if present,
    will be replaced with a tile's actual index when it is requested.

    If the value is an array: multiple URL templates. Each endpoint must return the same
    content for the same tile index. This can be used to work around domain sharding,
    allowing browsers to download more resources simultaneously. Requests made are
    balanced among the endpoints, based on the tile index.
    """

    tile_size = t.Int(None, allow_none=True).tag(sync=True)
    """
    The pixel dimension of the tiles, usually a power of 2.

    Tile size represents the target pixel width and height of each tile when rendered.
    Smaller tile sizes display the content at higher resolution, while the layer needs
    to load more tiles to fill the same viewport.

    - Type: `int`, optional
    - Default: `512`
    """

    zoom_offset = t.Int(None, allow_none=True).tag(sync=True)
    """
    This offset changes the zoom level at which the tiles are fetched. Needs to be an
    integer.

    - Type: `int`, optional
    - Default: `0`
    """

    max_zoom = t.Int(None, allow_none=True).tag(sync=True)
    """
    The max zoom level of the layer's data. When overzoomed (i.e. `zoom > max_zoom`),
    tiles from this level will be displayed.

    - Type: `int`, optional
    - Default: `None`
    """

    min_zoom = t.Int(None, allow_none=True).tag(sync=True)
    """
    The min zoom level of the layer's data. When underzoomed (i.e. `zoom < min_zoom`),
    the layer will not display any tiles unless `extent` is defined, to avoid issuing
    too many tile requests.

    - Type: `int`, optional
    - Default: `None`
    """

    extent = VariableLengthTuple(
        t.Float(),
        minlen=4,
        maxlen=4,
        allow_none=True,
        default_value=None,
    ).tag(sync=True)
    """
    The bounding box of the layer's data, in the form of `[min_x, min_y, max_x, max_y]`.
    If provided, the layer will only load and render the tiles that are needed to fill
    this box.

    - Type: `List[float]`, optional
    - Default: `None`
    """

    max_cache_size = t.Int(None, allow_none=True).tag(sync=True)
    """
    The maximum number of tiles that can be cached. The tile cache keeps loaded tiles in
    memory even if they are no longer visible. It reduces the need to re-download the
    same data over and over again when the user pan/zooms around the map, providing a
    smoother experience.

    If not supplied, the `max_cache_size` is calculated as 5 times the number of tiles
    in the current viewport.

    - Type: `int`, optional
    - Default: `None`
    """

    # TODO: Not sure if `getTileData` returns a `byteLength`?
    # max_cache_byte_size = t.Int(None, allow_none=True).tag(sync=True)
    # """
    # """

    refinement_strategy = t.Unicode(None, allow_none=True).tag(sync=True)
    """How the tile layer refines the visibility of tiles.

    When zooming in and out, if the layer only shows tiles from the current zoom level,
    then the user may observe undesirable flashing while new data is loading. By setting
    `refinement_strategy` the layer can attempt to maintain visual continuity by
    displaying cached data from a different zoom level before data is available.

    This prop accepts one of the following:

    - `"best-available"`: If a tile in the current viewport is waiting for its data to
      load, use cached content from the closest zoom level to fill the empty space. This
      approach minimizes the visual flashing due to missing content.
    - `"no-overlap"`: Avoid showing overlapping tiles when backfilling with cached
      content. This is usually favorable when tiles do not have opaque backgrounds.
    - `"never"`: Do not display any tile that is not selected.

    - Type: `str`, optional
    - Default: `"best-available"`
    """

    max_requests = t.Int(None, allow_none=True).tag(sync=True)
    """The maximum number of concurrent data fetches.

    If <= 0, no throttling will occur, and `get_tile_data` may be called an unlimited
    number of times concurrently regardless of how long that tile is or was visible.

    If > 0, a maximum of `max_requests` instances of `get_tile_data` will be called
    concurrently. Requests may never be called if the tile wasn't visible long enough to
    be scheduled and started. Requests may also be aborted (through the signal passed to
    `get_tile_data`) if there are more than `max_requests` ongoing requests and some of
    those are for tiles that are no longer visible.

    If `get_tile_data` makes fetch requests against an HTTP 1 web server, then
    max_requests should correlate to the browser's maximum number of concurrent fetch
    requests. For Chrome, the max is 6 per domain. If you use the data prop and specify
    multiple domains, you can increase this limit. For example, with Chrome and 3
    domains specified, you can set max_requests=18.

    If the web server supports HTTP/2 (Open Chrome dev tools and look for "h2" in the
    Protocol column), then you can make an unlimited number of concurrent requests (and
    can set max_requests=-1). Note that this will request data for every tile, no matter
    how long the tile was visible, and may increase server load.
    """

    desaturate = t.Float(0, min=0, max=1).tag(sync=True)
    """The desaturation of the bitmap. Between `[0, 1]`.

    - Type: `float`, optional
    - Default: `0`
    """

    transparent_color = VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to use for transparent pixels, in `[r, g, b, a]`.

    - Type: `List[float]`, optional
    - Default: `[0, 0, 0, 0]`
    """

    tint_color = VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to tint the bitmap by, in `[r, g, b]`.

    - Type: `List[float]`, optional
    - Default: `[255, 255, 255]`
    """


class ColumnLayer(BaseArrowLayer):
    """The ColumnLayer renders extruded cylinders (tessellated regular polygons) at given
    coordinates.
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[ColumnLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[ColumnLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[ColumnLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("column").tag(sync=True)

    table = ArrowTableTrait(allowed_geometry_types={EXTENSION_NAME.POINT})
    """A GeoArrow table with a Point or MultiPoint column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.ScatterplotLayer.from_geopandas] instead.
    """

    disk_resolution = t.Int(None, allow_none=True).tag(sync=True)
    """
    The number of sides to render the disk as. The disk is a regular polygon that fits
    inside the given radius. A higher resolution will yield a smoother look close-up,
    but also need more resources to render.

    - Type: `int`, optional
    - Default: `20`
    """

    radius = t.Float(None, allow_none=True).tag(sync=True)
    """
    Disk size in units specified by `radius_units` (default meters).

    - Type: `float`, optional
    - Default: `1000`
    """

    angle = t.Float(None, allow_none=True).tag(sync=True)
    """
    Disk rotation, counter-clockwise in degrees.

    - Type: `float`, optional
    - Default: `0`
    """

    offset = t.Tuple(t.Float(), t.Float(), default_value=None, allow_none=True).tag(
        sync=True,
    )
    """
    Disk offset from the position, relative to the radius. By default, the disk is
    centered at each position.

    - Type: `tuple[float, float]`, optional
    - Default: `(0, 0)`
    """

    coverage = t.Float(None, allow_none=True).tag(sync=True)
    """
    Radius multiplier, between 0 - 1. The radius of the disk is calculated by
    `coverage * radius`

    - Type: `float`, optional
    - Default: `1`
    """

    elevation_scale = t.Float(None, allow_none=True).tag(sync=True)
    """
    Column elevation multiplier. The elevation of column is calculated by
    `elevation_scale * get_elevation(d)`. `elevation_scale` is a handy property
    to scale all column elevations without updating the data.

    - Type: `float`, optional
    - Default: `1`
    """

    filled = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to draw a filled column (solid fill).

    - Type: `bool`, optional
    - Default: `True`
    """

    stroked = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to draw an outline around the disks. Only applies if `extruded=False`.

    - Type: `bool`, optional
    - Default: `False`
    """

    extruded = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to extrude the columns. If set to `false`, all columns will be rendered as
    flat polygons.

    - Type: `bool`, optional
    - Default: `True`
    """

    wireframe = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to generate a line wireframe of the column. The outline will have
    "horizontal" lines closing the top and bottom polygons and a vertical line
    (a "strut") for each vertex around the disk. Only applies if `extruded=True`.

    - Type: `bool`, optional
    - Default: `False`
    """

    flat_shading = t.Bool(None, allow_none=True).tag(sync=True)
    """
    If `True`, the vertical surfaces of the columns use [flat
    shading](https://en.wikipedia.org/wiki/Shading#Flat_vs._smooth_shading). If `false`,
    use smooth shading. Only effective if `extruded` is `True`.

    - Type: `bool`, optional
    - Default: `False`
    """

    radius_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the radius, one of `'meters'`, `'common'`, and `'pixels'`. See [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    line_width_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    line_width_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The line width multiplier that multiplied to all outlines if the `stroked` attribute
    is `True`.

    - Type: `float`, optional
    - Default: `1`
    """

    line_width_min_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The minimum outline width in pixels. This can be used to prevent the line from
    getting too small when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    line_width_max_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The maximum outline width in pixels. This can be used to prevent the line from
    getting too big when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    get_fill_color = ColorAccessor(None, allow_none=True)
    """
    The filled color of each object in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the filled color for
          all objects.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the filled color for the object at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_color = ColorAccessor(None, allow_none=True)
    """
    The outline color of each object in the format of `[r, g, b, [a]]`. Each channel is
    a number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the outline color for
          all objects.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the outline color for the object at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_elevation = FloatAccessor(None, allow_none=True)
    """
    The elevation of each cell in meters.

    Only applies if `extruded=True`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all polygons.
        - If an array is provided, each value in the array will be used as the width for
          the polygon at the same row index.
    - Default: `1000`.
    """

    get_line_width = FloatAccessor(None, allow_none=True)
    """
    The width of the outline of each column, in units specified by `line_width_units`
    (default `'meters'`). Only applies if `extruded: false` and `stroked: true`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the outline width for all columns.
        - If an array is provided, each value in the array will be used as the outline
          width for the column at the same row index.
    - Default: `1`.
    """


class PolygonLayer(BaseArrowLayer):
    """The `PolygonLayer` renders filled, stroked and/or extruded polygons.

    !!! note

        This layer is essentially a combination of a [`PathLayer`][lonboard.PathLayer]
        and a [`SolidPolygonLayer`][lonboard.SolidPolygonLayer]. This has some overhead
        beyond a `SolidPolygonLayer`, so if you're looking for the maximum performance
        with large data, you may want to use a `SolidPolygonLayer` directly.

    **Example:**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, PolygonLayer

    # A GeoDataFrame with Polygon or MultiPolygon geometries
    gdf = gpd.GeoDataFrame()
    layer = PolygonLayer.from_geopandas(
        gdf,
        get_fill_color=[255, 0, 0],
        get_line_color=[0, 100, 100, 150],
    )
    m = Map(layer)
    ```

    From [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest):

    ```py
    from geoarrow.rust.io import read_parquet
    from lonboard import Map, PolygonLayer

    # Example: A GeoParquet file with Polygon or MultiPolygon geometries
    table = read_parquet("path/to/file.parquet")
    layer = PolygonLayer(
        table=table,
        get_fill_color=[255, 0, 0],
        get_line_color=[0, 100, 100, 150],
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[PolygonLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[PolygonLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[PolygonLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("polygon").tag(sync=True)

    table = ArrowTableTrait(
        allowed_geometry_types={EXTENSION_NAME.POLYGON, EXTENSION_NAME.MULTIPOLYGON},
    )
    """A GeoArrow table with a Polygon or MultiPolygon column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.PolygonLayer.from_geopandas] instead.
    """

    stroked = t.Bool(None, allow_none=True).tag(sync=True)
    """Whether to draw an outline around the polygon (solid fill).

    Note that both the outer polygon as well the outlines of any holes will be drawn.

    - Type: `bool`, optional
    - Default: `True`
    """

    filled = t.Bool(None, allow_none=True).tag(sync=True)
    """Whether to draw a filled polygon (solid fill).

    Note that only the area between the outer polygon and any holes will be filled.

    - Type: `bool`, optional
    - Default: `True`
    """

    extruded = t.Bool(None, allow_none=True).tag(sync=True)
    """Whether to extrude the polygons.

    Based on the elevations provided by the `getElevation` accessor.

    If set to `false`, all polygons will be flat, this generates less geometry and is
    faster than simply returning 0 from getElevation.

    - Type: `bool`, optional
    - Default: `False`
    """

    wireframe = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to generate a line wireframe of the polygon. The outline will have
    "horizontal" lines closing the top and bottom polygons and a vertical line
    (a "strut") for each vertex on the polygon.

    - Type: `bool`, optional
    - Default: `False`

    **Remarks:**

    - These lines are rendered with `GL.LINE` and will thus always be 1 pixel wide.
    - Wireframe and solid extrusions are exclusive, you'll need to create two layers
      with the same data if you want a combined rendering effect.
    """

    elevation_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """Elevation multiplier.

    The final elevation is calculated by `elevationScale * getElevation(d)`.
    `elevationScale` is a handy property to scale all elevation without updating the
    data.

    - Type: `float`, optional
    - Default: `1`
    """

    line_width_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the outline width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    line_width_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The outline width multiplier that multiplied to all outlines of `Polygon` and
    `MultiPolygon` features if the `stroked` attribute is true.

    - Type: `float`, optional
    - Default: `1`
    """

    line_width_min_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The minimum outline width in pixels. This can be used to prevent the outline from
    getting too small when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    line_width_max_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The maximum outline width in pixels. This can be used to prevent the outline from
    getting too big when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    line_joint_rounded = t.Bool(None, allow_none=True).tag(sync=True)
    """Type of joint. If `true`, draw round joints. Otherwise draw miter joints.

    - Type: `bool`, optional
    - Default: `False`
    """

    line_miter_limit = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """The maximum extent of a joint in ratio to the stroke width.

    Only works if `line_joint_rounded` is false.

    - Type: `float`, optional
    - Default: `4`
    """

    get_fill_color = ColorAccessor(None, allow_none=True)
    """
    The fill color of each polygon in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the fill color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the fill color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_color = ColorAccessor(None, allow_none=True)
    """
    The outline color of each polygon in the format of `[r, g, b, [a]]`. Each channel is
    a number between 0-255 and `a` is 255 if not supplied.

    Only applies if `stroked=True`.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the outline color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the outline color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_width = FloatAccessor(None, allow_none=True)
    """
    The width of the outline of each polygon, in units specified by `line_width_units`
    (default `'meters'`).

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the outline width for all polygons.
        - If an array is provided, each value in the array will be used as the outline
          width for the polygon at the same row index.
    - Default: `1`.
    """

    get_elevation = FloatAccessor(None, allow_none=True)
    """
    The elevation to extrude each polygon with, in meters.

    Only applies if `extruded=True`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all polygons.
        - If an array is provided, each value in the array will be used as the width for
          the polygon at the same row index.
    - Default: `1000`.
    """


class ScatterplotLayer(BaseArrowLayer):
    """The `ScatterplotLayer` renders circles at given coordinates.

    **Example:**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, ScatterplotLayer

    # A GeoDataFrame with Point or MultiPoint geometries
    gdf = gpd.GeoDataFrame()
    layer = ScatterplotLayer.from_geopandas(
        gdf,
        get_fill_color=[255, 0, 0],
    )
    m = Map(layer)
    ```

    From [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest):

    ```py
    from geoarrow.rust.io import read_parquet
    from lonboard import Map, ScatterplotLayer

    # Example: A GeoParquet file with Point or MultiPoint geometries
    table = read_parquet("path/to/file.parquet")
    layer = ScatterplotLayer(
        table=table,
        get_fill_color=[255, 0, 0],
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[ScatterplotLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[ScatterplotLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[ScatterplotLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("scatterplot").tag(sync=True)

    table = ArrowTableTrait(
        allowed_geometry_types={EXTENSION_NAME.POINT, EXTENSION_NAME.MULTIPOINT},
    )
    """A GeoArrow table with a Point or MultiPoint column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.ScatterplotLayer.from_geopandas] instead.
    """

    radius_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the radius, one of `'meters'`, `'common'`, and `'pixels'`. See [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    radius_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    A global radius multiplier for all points.

    - Type: `float`, optional
    - Default: `1`
    """

    radius_min_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The minimum radius in pixels. This can be used to prevent the circle from getting
    too small when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    radius_max_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The maximum radius in pixels. This can be used to prevent the circle from getting
    too big when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    line_width_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    line_width_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    A global line width multiplier for all points.

    - Type: `float`, optional
    - Default: `1`
    """

    line_width_min_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The minimum line width in pixels. This can be used to prevent the stroke from
    getting too thin when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    line_width_max_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The maximum line width in pixels. This can be used to prevent the stroke from
    getting too thick when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    stroked = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Draw the outline of points.

    - Type: `bool`, optional
    - Default: `False`
    """

    filled = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Draw the filled area of points.

    - Type: `bool`, optional
    - Default: `True`
    """

    billboard = t.Bool(None, allow_none=True).tag(sync=True)
    """
    If `True`, rendered circles always face the camera. If `False` circles face up (i.e.
    are parallel with the ground plane).

    - Type: `bool`, optional
    - Default: `False`
    """

    antialiasing = t.Bool(None, allow_none=True).tag(sync=True)
    """
    If `True`, circles are rendered with smoothed edges. If `False`, circles are
    rendered with rough edges. Antialiasing can cause artifacts on edges of overlapping
    circles.

    - Type: `bool`, optional
    - Default: `True`
    """

    get_radius = FloatAccessor(None, allow_none=True)
    """
    The radius of each object, in units specified by `radius_units` (default
    `'meters'`).

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the radius for all objects.
        - If an array is provided, each value in the array will be used as the radius
          for the object at the same row index.
    - Default: `1`.
    """

    get_fill_color = ColorAccessor(None, allow_none=True)
    """
    The filled color of each object in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the filled color for
          all objects.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the filled color for the object at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_color = ColorAccessor(None, allow_none=True)
    """
    The outline color of each object in the format of `[r, g, b, [a]]`. Each channel is
    a number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the outline color
          for all objects.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the outline color for the object at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_width = FloatAccessor(None, allow_none=True)
    """
    The width of the outline of each object, in units specified by `line_width_units`
    (default `'meters'`).

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the outline width for all objects.
        - If an array is provided, each value in the array will be used as the outline
          width for the object at the same row index.
    - Default: `1`.
    """


class PathLayer(BaseArrowLayer):
    """The `PathLayer` renders lists of coordinate points as extruded polylines with
    mitering.

    **Example:**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, PathLayer

    # A GeoDataFrame with LineString or MultiLineString geometries
    gdf = gpd.GeoDataFrame()
    layer = PathLayer.from_geopandas(
        gdf,
        get_color=[255, 0, 0],
        width_min_pixels=2,
    )
    m = Map(layer)
    ```

    From [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest):

    ```py
    from geoarrow.rust.io import read_parquet
    from lonboard import Map, PathLayer

    # Example: A GeoParquet file with LineString or MultiLineString geometries
    table = read_parquet("path/to/file.parquet")
    layer = PathLayer(
        table=table,
        get_color=[255, 0, 0],
        width_min_pixels=2,
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[PathLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[PathLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[PathLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("path").tag(sync=True)

    table = ArrowTableTrait(
        allowed_geometry_types={
            EXTENSION_NAME.LINESTRING,
            EXTENSION_NAME.MULTILINESTRING,
        },
    )
    """A GeoArrow table with a LineString or MultiLineString column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.PathLayer.from_geopandas] instead.
    """

    width_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    width_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The path width multiplier that multiplied to all paths.

    - Type: `float`, optional
    - Default: `1`
    """

    width_min_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The minimum path width in pixels. This prop can be used to prevent the path from
    getting too thin when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    width_max_pixels = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    The maximum path width in pixels. This prop can be used to prevent the path from
    getting too thick when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    joint_rounded = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Type of joint. If `True`, draw round joints. Otherwise draw miter joints.

    - Type: `bool`, optional
    - Default: `False`
    """

    cap_rounded = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Type of caps. If `True`, draw round caps. Otherwise draw square caps.

    - Type: `bool`, optional
    - Default: `False`
    """

    miter_limit = t.Int(None, allow_none=True).tag(sync=True)
    """
    The maximum extent of a joint in ratio to the stroke width.
    Only works if `jointRounded` is `False`.

    - Type: `float`, optional
    - Default: `4`
    """

    billboard = t.Bool(None, allow_none=True).tag(sync=True)
    """
    If `True`, extrude the path in screen space (width always faces the camera).
    If `False`, the width always faces up.

    - Type: `bool`, optional
    - Default: `False`
    """

    get_color = ColorAccessor(None, allow_none=True)
    """
    The color of each path in the format of `[r, g, b, [a]]`. Each channel is a number
    between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the color for all
          paths.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the color for the path at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_width = FloatAccessor(None, allow_none=True)
    """
    The width of each path, in units specified by `width_units` (default `'meters'`).

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all paths.
        - If an array is provided, each value in the array will be used as the width for
          the path at the same row index.
    - Default: `1`.
    """


class PointCloudLayer(BaseArrowLayer):
    """The `PointCloudLayer` renders a point cloud with 3D positions, normals and colors.

    The `PointCloudLayer` can be more efficient at rendering large quantities of points
    than the [`ScatterplotLayer`][lonboard.ScatterplotLayer], but has fewer rendering
    options. In particular, you can have only one point size across all points in your
    data.

    **Example:**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, PointCloudLayer

    # A GeoDataFrame with Point geometries
    gdf = gpd.GeoDataFrame()
    layer = PointCloudLayer.from_geopandas(
        gdf,
        get_color=[255, 0, 0],
        point_size=2,
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[PointCloudLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[PointCloudLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[PointCloudLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("point-cloud").tag(sync=True)

    table = ArrowTableTrait(
        allowed_geometry_types={EXTENSION_NAME.POINT},
        allowed_dimensions={3},
    )
    """A GeoArrow table with a Point column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.PointCloudLayer.from_geopandas] instead.
    """

    size_units = t.Unicode(None, allow_none=True).tag(sync=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'pixels'`
    """

    point_size = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    Global radius of all points, in units specified by `size_units`.

    - Type: `float`, optional
    - Default: `10`
    """

    get_color = ColorAccessor(None, allow_none=True)
    """
    The color of each path in the format of `[r, g, b, [a]]`. Each channel is a number
    between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the color for all
          paths.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the color for the path at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_normal = NormalAccessor(None, allow_none=True)
    """
    The normal of each object, in `[nx, ny, nz]`.

    - Type: [NormalAccessor][lonboard.traits.NormalAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the normal for all
          points.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the normal for the point at the same row index.
    - Default: `1`.
    """


class SolidPolygonLayer(BaseArrowLayer):
    """The `SolidPolygonLayer` renders filled and/or extruded polygons.

    !!! note

        This layer is similar to the [`PolygonLayer`][lonboard.PolygonLayer] but will
        not render an outline around polygons. In most cases, you'll want to use the
        `PolygonLayer` directly, but for very large datasets not drawing the outline can
        significantly improve performance, in which case you may want to use this layer.

    **Example:**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, SolidPolygonLayer

    # A GeoDataFrame with Polygon or MultiPolygon geometries
    gdf = gpd.GeoDataFrame()
    layer = SolidPolygonLayer.from_geopandas(
        gdf,
        get_fill_color=[255, 0, 0],
    )
    m = Map(layer)
    ```

    From [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest):

    ```py
    from geoarrow.rust.io import read_parquet
    from lonboard import Map, SolidPolygonLayer

    # Example: A GeoParquet file with Polygon or MultiPolygon geometries
    table = read_parquet("path/to/file.parquet")
    layer = SolidPolygonLayer(
        table=table,
        get_fill_color=[255, 0, 0],
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[SolidPolygonLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[SolidPolygonLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[SolidPolygonLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("solid-polygon").tag(sync=True)

    table = ArrowTableTrait(
        allowed_geometry_types={EXTENSION_NAME.POLYGON, EXTENSION_NAME.MULTIPOLYGON},
    )
    """A GeoArrow table with a Polygon or MultiPolygon column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.SolidPolygonLayer.from_geopandas] instead.
    """

    filled = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to fill the polygons (based on the color provided by the
    `get_fill_color` accessor).

    - Type: `bool`, optional
    - Default: `True`
    """

    extruded = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to extrude the polygons (based on the elevations provided by the
    `get_elevation` accessor'). If set to `False`, all polygons will be flat, this
    generates less geometry and is faster than simply returning `0` from
    `get_elevation`.

    - Type: `bool`, optional
    - Default: `False`
    """

    wireframe = t.Bool(None, allow_none=True).tag(sync=True)
    """
    Whether to generate a line wireframe of the polygon. The outline will have
    "horizontal" lines closing the top and bottom polygons and a vertical line
    (a "strut") for each vertex on the polygon.

    - Type: `bool`, optional
    - Default: `False`

    **Remarks:**

    - These lines are rendered with `GL.LINE` and will thus always be 1 pixel wide.
    - Wireframe and solid extrusions are exclusive, you'll need to create two layers
      with the same data if you want a combined rendering effect.
    """

    elevation_scale = t.Float(None, allow_none=True, min=0).tag(sync=True)
    """
    Elevation multiplier. The final elevation is calculated by `elevation_scale *
    get_elevation(d)`. `elevation_scale` is a handy property to scale all elevation
    without updating the data.

    - Type: `float`, optional
    - Default: `1`
    """

    get_elevation = FloatAccessor(None, allow_none=True)
    """
    The elevation to extrude each polygon with, in meters.

    Only applies if `extruded=True`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all polygons.
        - If an array is provided, each value in the array will be used as the width for
          the polygon at the same row index.
    - Default: `1000`.
    """

    get_fill_color = ColorAccessor(None, allow_none=True)
    """
    The fill color of each polygon in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the fill color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the fill color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_color = ColorAccessor(None, allow_none=True)
    """
    The line color of each polygon in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    Only applies if `extruded=True`.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the line color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the line color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """


class HeatmapLayer(BaseArrowLayer):
    """The `HeatmapLayer` visualizes the spatial distribution of data.

    **Example**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, HeatmapLayer

    # A GeoDataFrame with Point geometries
    gdf = gpd.GeoDataFrame()
    layer = HeatmapLayer.from_geopandas(gdf)
    m = Map(layer)
    ```

    From [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest):

    ```py
    from geoarrow.rust.io import read_parquet
    from lonboard import Map, HeatmapLayer

    # Example: A GeoParquet file with Point geometries
    table = read_parquet("path/to/file.parquet")
    layer = HeatmapLayer(
        table=table,
        get_fill_color=[255, 0, 0],
    )
    m = Map(layer)
    ```

    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        **kwargs: Unpack[HeatmapLayerKwargs],
    ) -> None:
        err_msg = """
        The `HeatmapLayer` is not currently working.

        As of Lonboard v0.10, Lonboard upgraded to version 9.0 of the underlying
        [deck.gl](https://deck.gl/) library. deck.gl [appears to have a
        bug](https://github.com/visgl/deck.gl/issues/8960#issuecomment-2284791644) with
        the HeatmapLayer in 9.0, that has not yet been fixed.

        Please temporarily downgrade to Lonboard v0.9 if you would like to use the
        `HeatmapLayer`.
        """
        warnings.warn(dedent(err_msg), UserWarning)

        # NOTE: we override the default for _rows_per_chunk because otherwise we render
        # one heatmap per _chunk_ not for the entire dataset.
        table_o3 = Table.from_arrow(table)
        super().__init__(table=table, _rows_per_chunk=len(table_o3), **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[HeatmapLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[HeatmapLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("heatmap").tag(sync=True)

    table = ArrowTableTrait(allowed_geometry_types={EXTENSION_NAME.POINT})
    """A GeoArrow table with a Point column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.HeatmapLayer.from_geopandas] instead.
    """

    radius_pixels = t.Float(None, allow_none=True).tag(sync=True)
    """Radius of the circle in pixels, to which the weight of an object is distributed.

    - Type: `float`, optional
    - Default: `30`
    """

    # TODO: stabilize ColormapTrait
    # color_range?: Color[];
    # """Specified as an array of colors [color1, color2, ...].

    # - Default: `6-class YlOrRd` - [colorbrewer](http://colorbrewer2.org/#type=sequential&scheme=YlOrRd&n=6)
    # """

    intensity = t.Float(None, allow_none=True).tag(sync=True)
    """
    Value that is multiplied with the total weight at a pixel to obtain the final
    weight.

    - Type: `float`, optional
    - Default: `1`
    """

    threshold = t.Float(None, allow_none=True, min=0, max=1).tag(sync=True)
    """Ratio of the fading weight to the max weight, between `0` and `1`.

    For example, `0.1` affects all pixels with weight under 10% of the max.

    Ignored when `color_domain` is specified.

    - Type: `float`, optional
    - Default: `0.05`
    """

    color_domain = VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=2,
        maxlen=2,
    ).tag(sync=True)
    # """
    # Controls how weight values are mapped to the `color_range`, as an array of two
    # numbers [`min_value`, `max_value`].

    # - Type: `(float, float)`, optional
    # - Default: `None`
    # """

    aggregation = t.Unicode(None, allow_none=True).tag(sync=True)
    """Defines the type of aggregation operation

    Valid values are 'SUM', 'MEAN'.

    - Type: `str`, optional
    - Default: `"SUM"`
    """

    weights_texture_size = t.Int(None, allow_none=True).tag(sync=True)
    """Specifies the size of weight texture.

    - Type: `int`, optional
    - Default: `2048`
    """

    debounce_timeout = t.Int(None, allow_none=True).tag(sync=True)
    """
    Interval in milliseconds during which changes to the viewport don't trigger
    aggregation.

    - Type: `int`, optional
    - Default: `500`
    """

    get_weight = FloatAccessor(None, allow_none=True)
    """The weight of each object.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the weight for all objects.
        - If an array is provided, each value in the array will be used as the weight
          for the object at the same row index.
    - Default: `1`.
    """



from __future__ import annotations

import os
from pathlib import Path

from lonboard._constants import Environment

try:
    import IPython
except ImportError:
    IPython = None


def detect_environment() -> Environment:  # noqa: PLR0911
    """Try to detect user's environment.

    This is ported from Plotly here:
    https://github.com/plotly/plotly.py/blob/38ded4acc91bd6c33fbe2c8090d629ba2215dce1/packages/python/plotly/plotly/io/_renderers.py#L446C1-L538C37
    under the MIT license.
    """
    if IPython and IPython.get_ipython():
        try:
            import google.colab  # noqa: F401

            return Environment.Colab  # noqa: TRY300
        except ImportError:
            pass

    if Path("/kaggle/input").exists():
        return Environment.Kaggle

    if "AZURE_NOTEBOOKS_HOST" in os.environ:
        return Environment.Azure

    if "VSCODE_PID" in os.environ:
        return Environment.Vscode

    if "NTERACT_EXE" in os.environ:
        return Environment.Nteract

    if "COCALC_PROJECT_ID" in os.environ:
        return Environment.Cocalc

    if "DATABRICKS_RUNTIME_VERSION" in os.environ:
        return Environment.Databricks

    if IPython.get_ipython().__class__.__name__ == "TerminalInteractiveShell":
        return Environment.IPythonTerminal

    return Environment.Unknown


ENVIRONMENT = detect_environment()


DEFAULT_HEIGHT = "500px"



# ruff: noqa: D101
# Missing docstring in public class


from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Literal, TypeAlias, Union

import numpy as np
from arro3.core.types import ArrowArrayExportable, ArrowStreamExportable
from numpy.typing import NDArray

if sys.version_info >= (3, 12):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd
    import pyarrow as pa

    from lonboard._base import BaseExtension


IntFloat: TypeAlias = int | float
Units = Literal["meters", "common", "pixels"]

ColorAccessorInput = Union[
    list[int],
    tuple[int, int, int],
    tuple[int, ...],
    str,
    NDArray[np.uint8],
    "pa.FixedSizeListArray",
    "pa.ChunkedArray",
    ArrowArrayExportable,
    ArrowStreamExportable,
]
FloatAccessorInput = Union[
    int,
    float,
    NDArray[np.number],
    "pd.Series",
    "pa.FloatingPointArray",
    "pa.ChunkedArray",
    ArrowArrayExportable,
    ArrowStreamExportable,
]
NormalAccessorInput = Union[
    list[int],
    tuple[int, int, int],
    tuple[int, ...],
    NDArray[np.floating],
    "pa.FixedSizeListArray",
    "pa.ChunkedArray",
    ArrowArrayExportable,
    ArrowStreamExportable,
]
TextAccessorInput = Union[
    str,
    NDArray[np.str_],
    "pd.Series",
    "pa.StringArray",
    "pa.LargeStringArray",
    "pa.ChunkedArray",
    ArrowArrayExportable,
    ArrowStreamExportable,
]


class BaseLayerKwargs(TypedDict, total=False):
    extensions: Sequence[BaseExtension]
    pickable: bool
    visible: bool
    opacity: IntFloat
    auto_highlight: bool


class BitmapLayerKwargs(BaseLayerKwargs, total=False):
    image: str
    bounds: (
        tuple[IntFloat, IntFloat, IntFloat, IntFloat]
        | tuple[
            tuple[IntFloat, IntFloat],
            tuple[IntFloat, IntFloat],
            tuple[IntFloat, IntFloat],
            tuple[IntFloat, IntFloat],
        ]
    )

    desaturate: IntFloat
    transparent_color: Sequence[IntFloat]
    tint_color: Sequence[IntFloat]


class BitmapTileLayerKwargs(BaseLayerKwargs, total=False):
    data: str | Sequence[str]
    tile_size: int
    zoom_offset: int
    max_zoom: int
    min_zoom: int
    extent: Sequence[IntFloat]
    max_cache_size: int
    # max_cache_byte_size: int  # noqa: ERA001
    refinement_strategy: Literal["best-available", "no-overlap", "never"]
    max_requests: int
    desaturate: IntFloat
    transparent_color: Sequence[IntFloat]
    tint_color: Sequence[IntFloat]


class ColumnLayerKwargs(BaseLayerKwargs, total=False):
    disk_resolution: int
    radius: IntFloat
    angle: IntFloat
    offset: tuple[IntFloat, IntFloat]
    coverage: IntFloat
    elevation_scale: IntFloat
    filled: bool
    stroked: bool
    extruded: bool
    wireframe: bool
    flat_shading: bool
    radius_units: Units
    line_width_units: Units
    line_width_scale: IntFloat
    line_width_min_pixels: IntFloat
    line_width_max_pixels: IntFloat
    get_fill_color: ColorAccessorInput
    get_line_color: ColorAccessorInput
    get_elevation: FloatAccessorInput
    get_line_width: FloatAccessorInput


class PathLayerKwargs(BaseLayerKwargs, total=False):
    width_units: Units
    width_scale: IntFloat
    width_min_pixels: IntFloat
    width_max_pixels: IntFloat
    joint_rounded: bool
    cap_rounded: bool
    miter_limit: int
    billboard: bool
    get_color: ColorAccessorInput
    get_width: FloatAccessorInput


class PointCloudLayerKwargs(BaseLayerKwargs, total=False):
    size_units: Units
    point_size: IntFloat
    get_color: ColorAccessorInput
    get_normal: NormalAccessorInput


class PolygonLayerKwargs(BaseLayerKwargs, total=False):
    stroked: bool
    filled: bool
    extruded: bool
    wireframe: bool
    elevation_scale: IntFloat
    line_width_units: Units
    line_width_scale: IntFloat
    line_width_min_pixels: IntFloat
    line_width_max_pixels: IntFloat
    line_joint_rounded: bool
    line_miter_limit: IntFloat
    get_fill_color: ColorAccessorInput
    get_line_color: ColorAccessorInput
    get_line_width: FloatAccessorInput
    get_elevation: FloatAccessorInput


class ScatterplotLayerKwargs(BaseLayerKwargs, total=False):
    radius_units: Units
    radius_scale: IntFloat
    radius_min_pixels: IntFloat
    radius_max_pixels: IntFloat
    line_width_units: Units
    line_width_scale: IntFloat
    line_width_min_pixels: IntFloat
    line_width_max_pixels: IntFloat
    stroked: bool
    filled: bool
    billboard: bool
    antialiasing: bool
    get_radius: FloatAccessorInput
    get_fill_color: ColorAccessorInput
    get_line_color: ColorAccessorInput
    get_line_width: FloatAccessorInput


class SolidPolygonLayerKwargs(BaseLayerKwargs, total=False):
    filled: bool
    extruded: bool
    wireframe: bool
    elevation_scale: IntFloat
    get_elevation: FloatAccessorInput
    get_fill_color: ColorAccessorInput
    get_line_color: ColorAccessorInput


class TripsLayerKwargs(BaseLayerKwargs, total=False):
    width_units: Units
    width_scale: IntFloat
    width_min_pixels: IntFloat
    width_max_pixels: IntFloat
    joint_rounded: bool
    cap_rounded: bool
    miter_limit: IntFloat
    billboard: bool
    fade_trail: bool
    trail_length: IntFloat
    current_time: IntFloat
    get_color: ColorAccessorInput
    get_width: FloatAccessorInput


class HeatmapLayerKwargs(BaseLayerKwargs, total=False):
    radius_pixels: IntFloat
    intensity: IntFloat
    threshold: IntFloat
    color_domain: tuple[IntFloat, IntFloat]
    aggregation: Literal["SUM", "MEAN"]
    weights_texture_size: int
    debounce_timeout: int
    get_weight: FloatAccessorInput
